local MODNAME = ...

-------------------------------------------------------------------------------
-- Localized Lua globals
-------------------------------------------------------------------------------
local _G = getfenv(0)

local assert = _G.assert
local ceil = _G.ceil
local concat = _G.table.concat
local floor = _G.floor
local format = _G.format
local min = _G.min
local tonumber = _G.tonumber
local tostring = _G.tostring
local tostringall = _G.tostringall
local type = _G.type

local ipairs = _G.ipairs
local pairs = _G.pairs
local rawset = _G.rawset
local select = _G.select
local setmetatable = _G.setmetatable
local sort = _G.sort
local tinsert = _G.tinsert
local tremove = _G.tremove
local unpack = _G.unpack
local wipe = _G.table.wipe


local LibStub = _G.LibStub


-- Fix for missing functions in MoP
local GetUnspentTalentPoints = _G.GetUnspentTalentPoints or _G.GetNumUnspentTalents or function() return 0 end
local GetNumTalentGroups = _G.GetNumTalentGroups or _G.GetNumSpecGroups or function() return 0 end

-------------------------------------------------------------------------------
-- AddOn namespace
-------------------------------------------------------------------------------
local Syllabus = LibStub("AceAddon-3.0"):NewAddon(MODNAME, "AceEvent-3.0", "AceConsole-3.0")
_G[MODNAME] = Syllabus

local LDB = _G.LibStub("LibDataBroker-1.1")
local LMD = _G.LibStub("LibMapData-1.0")
local QTip = _G.LibStub("LibQTip-1.0")

local dev = false
--[===[@debug@
dev = true
--@end-debug@]===]

local alpha = false
--@alpha@
alpha = true
--@end-alpha@

local L = _G.LibStub("AceLocale-3.0"):GetLocale(MODNAME, "enUS", true, dev)

local DataObj
--local tooltip

local version = _G.GetAddOnMetadata(MODNAME, "Version")
version = dev and (version .. " - Development") or (alpha and version .. " - Alpha") or version
Syllabus.version = version

-------------------------------------------------------------------------------
-- Compatibility with Mists of Pandaria
-------------------------------------------------------------------------------

local IS_504 = select(4, _G.GetBuildInfo()) >= 50000
local IS_MOP = min(_G.GetExpansionLevel(),_G.GetAccountExpansionLevel()) >= 4


-------------------------------------------------------------------------------
-- Functions
-------------------------------------------------------------------------------

-- print functions
function Syllabus:Print(msg)
	_G.SELECTED_CHAT_FRAME:AddMessage(MODNAME .. ": " .. msg, 0.5, 1, 0)
end

function Syllabus:debug(msg, ...)
	if dev then
		self:err(msg, ...)
	end
end

function Syllabus:err(msg,...) _G.geterrorhandler()(msg:format(tostringall(...)) .. " - " .. _G.time()) end

local function print(...) Syllabus:Print(...) end
local function debug(...) Syllabus:debug(...) end
local function err(...) Syllabus:err(...) end

-- Table copy function -- borrowed from AceDB-3.0.lua
local function CopyTable(src, dest)
	if type(dest) ~= "table" then dest = {} end
	if type(src) == "table" then
		for k,v in pairs(src) do
			if type(v) == "table" then
				-- try to index the key first so that the metatable creates the defaults, if set, and use that table
				v = CopyTable(v, dest[k])
			end
			dest[k] = v
		end
	end
	return dest
end

-- Table reuse functions
local AcquireTable, ReleaseTable
do
	local table_cache = {}

	-- Returns a table
	function AcquireTable()
		local tbl = tremove(table_cache) or {}
		return tbl
	end

	-- Cleans the table and stores it in the cache
	function ReleaseTable(tbl)
		if not tbl then return end

		-- Nested tables ?
		for i=1,#tbl do
			if type(tbl[i]) == "table" then ReleaseTable(tbl[i]) end
		end

		wipe(tbl)
		tinsert(table_cache, tbl)
	end
end	-- do block

-- Check for empty table
local function IsTableEmpty(table)
	return not _G.next(table)
end

-------------------------------------------------------------------------------
-- Constants
-------------------------------------------------------------------------------

local MY_CLASS, MY_CLASS_UPPER = _G.UnitClass("player")
local MY_LEVEL = _G.UnitLevel("player")
local MY_NAME = _G.UnitName("player")
local MY_REALM = _G.GetRealmName()
local MY_RACE, MY_RACE_EN = _G.UnitRace("player")
local MY_FACTION = _G.UnitFactionGroup("player")
local MY_FACTION_FLAG = (MY_FACTION == 'Alliance' and 'is_alliance') or 'is_horde'

local COLOR_TABLE = _G.CUSTOM_CLASS_COLORS or _G.RAID_CLASS_COLORS
local CLASS_COLORS = {}

local MY_PROFESSIONS = {}
Syllabus.prof = MY_PROFESSIONS

for k, v in pairs(COLOR_TABLE) do
	CLASS_COLORS[k] = ("%2x%2x%2x"):format(v.r * 255, v.g * 255, v.b * 255)
end

-- Get the localized continent names
local CONTINENT_NAMES = { _G.GetMapContinents() }

-- Get the localized zone names
local ZONE_NAMES = {}
for continent_id = 1,#CONTINENT_NAMES do
	ZONE_NAMES[continent_id] = { _G.GetMapZones(continent_id) }
end
-- Special cases for zones not included in GetMapZones()
ZONE_NAMES[-1]		= {}
ZONE_NAMES[-1][0]	= _G.UNKNOWN
ZONE_NAMES[0]		= {}
ZONE_NAMES[0][0]	= _G.UNKNOWN

local MAPID_TO_NAME = {}
-- Cache names locally
setmetatable(MAPID_TO_NAME, { __index = function(t, k)
	if k == nil then return end
	local v = LMD:MapLocalize(k)
	rawset(t, k, v)
	return v
end })


-- Profession titles
-- The key is the max rank for a skill level, the number is the SpellID where we can get the title name
local TITLES = {
	[_G.GetSpellInfo(2575)] = {	-- Mining
		[75] = 2575,
		[150] = 2576,
		[225] = 3564,
		[300] = 10248,
		[375] = 29354,
		[450] = 50310,
		[525] = 74517,
	},
	[_G.GetSpellInfo(131474)] = {	-- Fishing
		[75] = 7733,
		[150] = 7734,
		[225] = 54083,
		[300] = 18249,
		[375] = 54084,
		[450] = 51293,
		[525] = 88869,
	},
}

if IS_MOP then
	TITLES[_G.GetSpellInfo(2575)][600] = 102161
	TITLES[_G.GetSpellInfo(131474)][600] = 110412
end

-- Professions that are not specifically listed in the spellbook
-- ex. Way of the ... are seen by the game as profession but not listed in the profession page
local EXTRA_PROFESSIONS = {}
if IS_MOP then
	tinsert(EXTRA_PROFESSIONS,125589) -- Way of the Brew
	tinsert(EXTRA_PROFESSIONS,125588) -- Way of the Oven
	tinsert(EXTRA_PROFESSIONS,125584) -- Way of the Wok
	tinsert(EXTRA_PROFESSIONS,124694) -- Way of the Gril
	tinsert(EXTRA_PROFESSIONS,125586) -- Way of the Pot
	tinsert(EXTRA_PROFESSIONS,125587) -- Way of the Steamer
end

local EXTRA_PROFESSIONS_NAMES = {}
for _, spell_id in ipairs(EXTRA_PROFESSIONS) do
	local name = _G.GetSpellInfo(spell_id)
	if not name then
		err("Missign profession name for %s",spell_id)
	else
		EXTRA_PROFESSIONS_NAMES[name] = spell_id
	end
end

-- Implied skills: knowing a skill imply that you know some other skills
-- The professions can probably be done that way for the TITLES
local IMPLIED_SKILLS = {
	[88961] = { 78670 }, 												-- Journeyman Archaeology => Apprentice Archaeology
	[89718] = { 88961, 78670 }, 										-- Expert Archaeology => Journeyman Archaeology
	[89719] = { 89718, 88961, 78670 }, 								-- Artisan Archaeology => Expert Archaeology
	[89720] = { 89719, 89718, 88961, 78670 }, 					-- Master Archaeology => Artisan Archaeology
	[89721] = { 89720, 89719, 89718, 88961, 78670 }, 			-- Grand Master Archaeology => Master Archaeology
	[89722] = { 89721, 89720, 89719, 89718, 88961, 78670 }, 	-- Illustrious Grand Master Archaeology => Grand Master Archaeology

	[3102]  = { 2550 }, 													-- Journeyman Cooking => Apprentice Cooking
	[3413]  = { 3102, 2550 }, 											-- Expert Cooking => Journeyman Cooking
	[18260] = { 3413, 3102, 2550 }, 									-- Artisan Cooking => Expert Cooking
	[33359] = { 18260, 3413, 3102, 2550 }, 						-- Master Cooking => Artisan Cooking
	[51296] = { 33359, 18260, 3413, 3102, 2550 }, 				-- Grand Master Cooking => Master Cooking
	[88053] = { 51296, 33359, 18260, 3413, 3102, 2550 }, 		-- Illustrious Grand Master Cooking => Grand Master Cooking

	[3274]  = { 3273 }, 													-- Journeyman First Aid => Apprentice First Aid
	[7924]  = { 3274, 3273 }, 											-- Expert First Aid => Journeyman First Aid
	[10846] = { 7924, 3274, 3273 }, 									-- Artisan First Aid => Expert First Aid
	[27028] = { 10846, 7924, 3274, 3273 }, 						-- Master First Aid => Artisan First Aid
	[45542] = { 27028, 10846, 7924, 3274, 3273 }, 				-- Grand Master First Aid => Master First Aid
	[74559] = { 45542, 27028, 10846, 7924, 3274, 3273 }, 		-- Illustrious Grand Master First Aid => Grand Master First Aid

	-- Fishing doesn't get listed in the Spellbook as such. We get 131474 that is the spell for fishing that has no rank 
--	[7731]  = { 7620 }, 													-- Journeyman Fishing => Apprentice Fishing
--	[7732]  = { 7731, 7620 }, 											-- Expert Fishing => Journeyman Fishing
--	[18248] = { 7732, 7731, 7620 }, 									-- Artisan Fishing => Expert Fishing
--	[33095] = { 18248, 7732, 7731, 7620 }, 						-- Master Fishing => Artisan Fishing
--	[51294] = { 33095, 18248, 7732, 7731, 7620 }, 				-- Grand Master Fishing => Master Fishing
--	[88868] = { 51294, 33095, 18248, 7732, 7731, 7620 }, 		-- Illustrious Grand Master Fishing => Grand Master Fishing

	[3101]  = { 2259 },													-- Journeyman Alchemy => Apprentice Alchemy
	[3464]  = { 3101, 2259 },											-- Expert Alchemy => Journeyman Alchemy
	[11611] = { 3464, 3101, 2259 },									-- Artisan Alchemy => Expert Alchemy
	[28596] = { 11611, 3464, 3101, 2259 },							-- Master Alchemy => Artisan Alchemy
	[51304] = { 28596, 11611, 3464, 3101, 2259 },				-- Grand Master Alchemy => Master Alchemy
	[80731] = { 51304, 28596, 11611, 3464, 3101, 2259 },		-- Illustrious Grand Master Alchemy => Grand Master Alchemy

	[3100]  = { 2018 },													-- Journeyman Blacksmithing => Apprentice Blacksmithing
	[3538]  = { 3100, 2018 },											-- Expert Blacksmithing => Journeyman Blacksmithing
	[9785]  = { 3538, 3100, 2018 },									-- Artisan Blacksmithing => Expert Blacksmithing
	[29844] = { 9785, 3538, 3100, 2018 },							-- Master Blacksmithing => Artisan Blacksmithing
	[51300] = { 29844, 9785, 3538, 3100, 2018 },					-- Grand Master Blacksmithing => Master Blacksmithing
	[76666] = { 51300, 29844, 9785, 3538, 3100, 2018 },		-- Illustrious Grand Master Blacksmithing => Grand Master Blacksmithing

	[7412]  = { 7411 },													-- Journeyman Enchanting => Apprentice Enchanting
	[7413]  = { 7412, 7411 },											-- Expert Enchanting => Journeyman Enchanting
	[13920] = { 7413, 7412, 7411 },									-- Artisan Enchanting => Expert Enchanting
	[28029] = { 13920, 7413, 7412, 7411 },							-- Master Enchanting => Artisan Enchanting
	[51313] = { 28029, 13920, 7413, 7412, 7411 },				-- Grand Master Enchanting => Master Enchanting
	[74258] = { 51313, 28029, 13920, 7413, 7412, 7411 },		-- Illustrious Grand Master Enchanting => Grand Master Enchanting

	[4037]  = { 4036 },													-- Grand Master Engineering => Master Engineering
	[4038]  = { 4037, 4036 },											-- Grand Master Engineering => Master Engineering
	[12656] = { 4038, 4037, 4036 },									-- Grand Master Engineering => Master Engineering
	[30350] = { 12656, 4038, 4037, 4036 },							-- Grand Master Engineering => Master Engineering
	[51306] = { 30350, 12656, 4038, 4037, 4036 },				-- Grand Master Engineering => Master Engineering
	[82774] = { 51306, 30350, 12656, 4038, 4037, 4036 },		-- Illustrious Grand Master Engineering => Grand Master Engineering

	[2368]  = { 2366 },													-- Journeyman Herbalism => Apprentice Herbalism
	[3570]  = { 2368, 2366 },											-- Expert Herbalism => Journeyman Herbalism
	[11993] = { 3570, 2368, 2366 },									-- Artisan Herbalism => Expert Herbalism
	[28695] = { 11993, 3570, 2368, 2366 },							-- Master Herbalism => Artisan Herbalism
	[50300] = { 28695, 11993, 3570, 2368, 2366 },				-- Grand Master Herbalism => Master Herbalism
	[74519] = { 50300, 28695, 11993, 3570, 2368, 2366 },		-- Illustrious Grand Master Herbalism => Grand Master Herbalism

	[45358] = { 45357 },													-- Journeyman Inscription => Apprentice Inscription
	[45359] = { 45358, 45357 },										-- Expert Inscription => Journeyman Inscription
	[45360] = { 45359, 45358, 45357 },								-- Artisan Inscription => Master Inscription
	[45361] = { 45360, 45359, 45358, 45357 },						-- Master Inscription => Artisan Inscription
	[45363] = { 45361, 45360, 45359, 45358, 45357 },			-- Grand Master Inscription => Master Inscription
	[86008] = { 45363, 45361, 45360, 45359, 45358, 45357 },	-- Illustrious Grand Master Inscription => Grand Master Inscription

	[25230] = { 25229 }, 												-- Journeyman Jewelcrafter => Apprentice Jewelcrafter
	[28894] = { 25230, 25229 }, 										-- Expert Jewelcrafter => Journeyman Jewelcrafter
	[28895] = { 28894, 25230, 25229 }, 								-- Artisan Jewelcrafter => Expert Jewelcrafter
	[28897] = { 28895, 28894, 25230, 25229 }, 					-- Master Jewelcrafter => Artisan Jewelcrafter
	[51311] = { 28897, 28895, 28894, 25230, 25229 },			-- Grand Master Jewelcrafter => Master Jewelcrafter
	[73318] = { 51311, 28897, 28895, 28894, 25230, 25229 },	-- Illustrious Grand Master Jewelcrafter => Grand Master Jewelcrafter

	[3104]  = { 2108 },													-- Journeyman Leatherworking => Apprentice Leatherworking
	[3811]  = { 3104, 2108 },											-- Expert Leatherworking => Journeyman Leatherworking
	[10662] = { 3811, 3104, 2108 },									-- Artisan Leatherworking => Expert Leatherworking
	[32549] = { 10662, 3811, 3104, 2108 },							-- Master Leatherworking => Artisan Leatherworking
	[51302] = { 32549, 10662, 3811, 3104, 2108 },				-- Grand Master Leatherworking => Master Leatherworking
	[81199] = { 51302, 32549, 10662, 3811, 3104, 2108 },		-- Illustrious Grand Master Leatherworking => Grand Master Leatherworking

	-- Mining doesn't get listed in the Spellbook, we get Smelthing without a rank
	--[2576]  = { 2575 },											-- Journeyman Mining => Apprentice Mining
	--[3564]  = { 2576, 2575 },									-- Expert Mining => Journeyman Mining
	--[10248] = { 3564, 2576, 2575 },							-- Artisan Mining => Expert Mining
	--[29354] = { 10248, 3564, 2576, 2575 },					-- Master Mining => Artisan Mining
	--[50310] = { 29354, 10248, 3564, 2576, 2575 },		-- Grand Master Mining => Master Mining

	[8617]  = { 8613 },													-- Journeyman Skinning => Apprentice Skinning
	[8618]  = { 8617, 8613 },											-- Expert Skinning => Journeyman Skinning
	[10768] = { 8618, 8617, 8613 },									-- Artisan Skinning => Expert Skinning
	[32678] = { 10768, 8618, 8617, 8613 },							-- Master Skinning => Artisan Skinning
	[50305] = { 32678, 10768, 8618, 8617, 8613 },				-- Grand Master Skinning => Master Skinning
	[74522] = { 50305, 32678, 10768, 8618, 8617, 8613 },		-- Illustrious Grand Master Skinning => Grand Master Skinning

	[3909]  = { 3908 },													-- Journeyman Tailoring => Apprentice Tailoring
	[3910]  = { 3909, 3908 },											-- Expert Tailoring => Journeyman Tailoring
	[12180] = { 3910, 3909, 3908 },									-- Artisan Tailoring => Expert Tailoring
	[26790] = { 12180, 3910, 3909, 3908 },							-- Master Tailoring => Artisan Tailoring
	[51309] = { 26790, 12180, 3910, 3909, 3908 },				-- Grand Master Tailoring => Master Tailoring
	[75156] = { 51309, 26790, 12180, 3910, 3909, 3908 },		-- Grand Master Tailoring => Master Tailoring

	[33391] = { 33388 }, 												-- Journeyman Riding => Apprentice Riding
	[34090] = { 33391, 33388 }, 										-- Expert Riding => Journeyman Riding
	[34091] = { 34090, 33391, 33388 }, 								-- Artisan Ridin => Expert Riding
	[54197] = { 34090, 33391, 33388 }, 								-- Cold Weather Flying => Expert Riding
	[90265] = { 34091, 34090, 33391, 33388 }, 					-- Master Riding => Artisan Riding

	[86530] = { 87505 }, 												-- Druid Leather specialization
	[86531] = { 87504 }, 												-- Rogue Leather specialization
	[86528] = { 87506, 8737 },											-- Hunter Mail specialization
	[86529] = { 87507, 8737 },											-- Shaman Mail specialization
	[86524] = { 87510 }, 												-- Death Knight Plate specialization
	[86525] = { 87511, 750 },											-- Paladin Plate Specialization
	[86526] = { 87509, 750 },											-- Warrior Plate Specialization

	[86471] = { 87492 },													-- Death Knight Mastery
	[86470] = { 87491 },													-- Druid Mastery
	[86472] = { 87493 },													-- Hunter Mastery
	[86473] = { 86467 },													-- Mage Mastery
	[86474] = { 87494 },													-- Paladin Mastery
	[86475] = { 87495 },													-- Priest Mastery
	[86476] = { 87496 },													-- Rogue Mastery
	[86477] = { 87497 },													-- Shaman Mastery
	[86478] = { 87498 }, 												-- Warlock Mastery
	[86479] = { 87500 },													-- Warrior Mastery

	[93321] = { 79682 },											-- Hunter Control Pet
	[93375] = { 80388 },											-- Warlock Control Demon
	[33943] = { 34090, 33391, 33388 },						-- Druid Flight Form => Expert Riding
	[40120] = { 33943, 34090, 33391, 33388 },				-- Swift Flight Form => Flight Form
}

if IS_MOP then
	-- Add the new level for Mists of Pandaria
	IMPLIED_SKILLS[110393] = { 89722, 89721, 89720, 89719, 89718, 88961, 78670 }			-- Archeology
	IMPLIED_SKILLS[104381] = { 88053, 51296, 33359, 18260, 3413, 3102, 2550 } 				-- Cooking
	IMPLIED_SKILLS[110406] = { 74559, 45542, 27028, 10846, 7924, 3274, 3273 } 				-- First Aid
--	IMPLIED_SKILLS[110412] = { 88868, 51294, 33095, 18248, 7732, 7731, 7620 } 				-- Fishing
	IMPLIED_SKILLS[105206] = { 80731, 51304, 28596, 11611, 3464, 3101, 2259 }				-- Alchemy
	IMPLIED_SKILLS[110396] = { 76666, 51300, 29844, 9785, 3538, 3100, 2018 }				-- Blacksmithing
	IMPLIED_SKILLS[110400] = { 74258, 51313, 28029, 13920, 7413, 7412, 7411 }				-- Enchanting
	IMPLIED_SKILLS[110403] = { 82774, 51306, 30350, 12656, 4038, 4037, 4036 }				-- Engineering
	IMPLIED_SKILLS[110413] = { 74519, 50300, 28695, 11993, 3570, 2368, 2366 }				-- Herbalism
	IMPLIED_SKILLS[110417] = { 86008, 45363, 45361, 45360, 45359, 45358, 45357 }			-- Inscription
	IMPLIED_SKILLS[110420] = { 73318, 51311, 28897, 28895, 28894, 25230, 25229 }			-- Jewelcrafter
	IMPLIED_SKILLS[110423] = { 81199, 51302, 32549, 10662, 3811, 3104, 2108 }				-- Leatherworking
	IMPLIED_SKILLS[102216] = { 74522, 50305, 32678, 10768, 8618, 8617, 8613 }				-- Skinning
	IMPLIED_SKILLS[110426] = { 75156, 51309, 26790, 12180, 3910, 3909, 3908 }				-- Tailoring
end

-- Abilities that are requirements for skills
local IS_ABILITY = {
	[71] 		= true, -- Defensive Stance
	[339] 	= true, -- Entangling Roots
	[527] 	= true, -- Dispel Magic
	[768] 	= true, -- Cat Form
	[1120] 	= true, -- Drain Soul
	[1515] 	= true, -- Tame Beast
	[1784] 	= true, -- Stealth
	[2458] 	= true, -- Berserker Stance
	[2836] 	= true, -- Detect Traps
	[5487] 	= true, -- Bear Form
	[20219] 	= true, -- Gnomish Engineer
	[20222] 	= true, -- Goblin Engineer
	[33388]	= true, -- Apprentice Riding
	[34090]  = true, -- Expert Riding
	[33391] 	= true, -- Journeyman Riding
	[34091] 	= true, -- Artisan Riding
	[42955] 	= true, -- Conjure Refreshment
	[55642]	= true, -- Lightweave Embroidery (Rank 1)
	[55769]	= true, -- Darkglow Embroidery (Rank 1)
	[55777]	= true, -- Swordguard Embroidery (Rank 1)
	[56034]	= true, -- Master's Spellthread (Rank 1)
	[56039]	= true, -- Sanctified Spellthread (Rank 1)
	[74558]  = true, -- Field Bandage: Dense Embersilk
}

-- Will be populated during initialization
local ABILITY_NAME = {} -- SpellID => Long Ability Name
-- Need to put meta for missing ability name here
local SHORT_ABILITY_NAME = {} -- Long Ability Name => Short Ability Name
local ABILITY_NAME_TO_SPELLID = {} -- Long Ability Name => SpellID

-- Skill that requires these ability will not be displayed unless the ability is known
-- Mostly, if you are not a Gnome Engineer, you don't see that you could learn Gnomish Shrink Ray and
-- the like if you were a Gnomish Engineer. Speciality just generated to much crap.
-- If someone has an idea for a better idea, let me know - LaoTseu
local IS_ABILITY_EXCEPTION = {
	[20219] 	= true, -- Gnomish Engineer
	[20222] 	= true, -- Goblin Engineer
}
local IS_ABILITY_EXCEPTION_NAME = {} -- Long Ability Name => SpellID, need it to test against a name
-- Debug
if dev then
	Syllabus.ABILITY_NAME = ABILITY_NAME
	Syllabus.SHORT_ABILITY_NAME = SHORT_ABILITY_NAME
	Syllabus.IS_ABILITY_EXCEPTION_NAME = IS_ABILITY_EXCEPTION_NAME
end


-- To find the spell_id using the texture
local TRACKING = {
	DRUID = {
		["Interface\\Icons\\Ability_Tracking"] = 5225, -- Track Humanoids
	},
	HUNTER = {
		["Interface\\Icons\\Ability_Racial_Avatar"] = 19882, -- Track Giants
		["Interface\\Icons\\INV_Misc_Head_Dragon_01"] = 19879, -- Track Dragonkin
		["Interface\\Icons\\Spell_Holy_PrayerOfHealing"] = 19883, -- Track Humanoids
		["Interface\\Icons\\Spell_Shadow_SummonFelHunter"] = 19878, -- Track Demons
		["Interface\\Icons\\Ability_Stealth"] = 19885, -- Track Hidden
		["Interface\\Icons\\Spell_Shadow_DarkSummoning"] = 19884, -- Track Undead
		["Interface\\Icons\\Ability_Tracking"] = 1494, -- Track Beasts
		["Interface\\Icons\\Spell_Frost_SummonWaterElemental"] = 19880, -- Track Elementals
	},
}

-- Special case for skills that require for the crafter to be able to wear a certain type of armor
local NEED_ARMOR = {
	[41320]		= 'Cloth', -- Destruction Holo-gogs
	[41321]		= 'PRIEST', -- Powerheal 4000 Lens
	[56484]		= 'Cloth', -- Visage Liquification Goggles
	[56465]		= 'Cloth', -- Mechanized Snow Goggles
	[81725]		= 'Cloth', -- Lightweight Bio-Optic Killshades
	[41317]		= 'Leather', -- Deathblow X11 Goggles
	[41318]		= 'Leather', -- Wonderheal XT40 Shades
	[41319]		= 'Leather', -- Magnified Moon Specs
	[56481]		= 'Leather', -- Weakness Spectralizers
	[56486]		= 'Leather', -- Greensight Gogs
	[61481]		= 'Leather', -- Mechanized Snow Goggles
	[81724]		= 'Leather', -- Camouflage Bio-Optic Killshades
	[81722]		= 'Leather', -- Agile Bio-Optic Killshades
	[41314]		= 'Mail', -- Surestrike Goggles v2.0
	[41315]		= 'Mail', -- Gadgetstorm Goggles
	[41316]		= 'Mail', -- Living Replicator Specs
	[56487]		= 'Mail', -- Electroflux Sight Enhancers
	[56574]		= 'Mail', -- Truesight Ice Blinders
	[61482]		= 'Mail', -- Mechanized Snow Goggles
	[81716]		= 'Mail', -- Deadly Bio-Optic Killshades
	[81720]		= 'Mail', -- Energized Bio-Optic Killshades
	[40274]		= 'Plate',	-- Furious Gizmatic Goggles
	[41311]		= 'PALADIN', -- Justicebringer 2000 Specs
	[41312]		= 'Plate', -- Tankatronic Goggles
	[56480]		= 'Plate', -- Armored Titanium Goggles
	[56483]		= 'Plate', -- Charged Titanium Specs
	[61483]		= 'Plate', -- Mechanized Snow Goggles
	[62271]		= 'PALADIN', -- Unbreakable Healing Amplifiers
	[81715]		= 'Plate', -- Specialized Bio-Optic Killshades
	[81714]		= 'Plate', -- Reinforced Bio-Optic Killshades
}

-- Type of armor that a class can wear at max level
local ARMOR_BY_CLASS = {
	MAGE			= 'Cloth',
	PRIEST		= 'Cloth',
	WARLOCK		= 'Cloth',
	DRUID			= 'Leather',
	MONK			= 'Leather',
	ROGUE			= 'Leather',
	HUNTER		= 'Mail',
	SHAMAN		= 'Mail',
	PALADIN		= 'Plate',
	WARRIOR		= 'Plate',
	DEATHKNIGHT	= 'Plate',
}

-- Shared skills - skills that are shared between all classes even though
-- IsTradeskillTrainer() return false
-- No longer that way but I like to keep them with the classe skill. Riding trainers now return IsTradeskillTrainer() == true
local RIDING_SKILLS = {
	[33388]	= true, -- Apprentice Riding
	[33391]	= true, -- Journeyman Riding
	[34090]	= true, -- Expert Riding
	[34091]	= true, -- Artisan Riding
	[54197]	= true, -- Cold Weather Flying
	[90265]	= true, -- Master Riding
	[90267]	= true, -- Flight Master's License
}

-- Racial exceptions
-- Some skills are only available to certain races
local SKILL_RACIAL_EXCEPTIONS = {
	[69820] = { "Tauren" }, -- Summon Sunwalker Kodo
	[69826] = { "Tauren" }, -- Summon Great Sunwalker Kodo
	[34769] = { "BloodElf", "Draenei", "Dwarf", "Human" }, -- Summon Warhorse
	[34767] = { "BloodElf", "Draenei", "Dwarf", "Human" }, -- Summon Charger
}

-------------------------------------------------------------------------------
-- Fonts
-------------------------------------------------------------------------------

-- Util function to create font of a certain color
local function CreateColorFont(name, r, g, b, font_type)
	font_type = font_type or _G.GameTooltipText
	local font = _G.CreateFont(MODNAME.."_"..name)
	--font:CopyFontObject(font_type or _G.GameTooltipText)

	local filename, size, flags = font_type:GetFont()
	font:SetFont(filename, size, flags)

	local sr, sg, sb, sa = font_type:GetShadowColor()
	font:SetShadowColor(sr, sg, sb, sa)

	font:SetTextColor(r, g, b)
	return font
end

local GOLD_FONT = CreateColorFont("GOLD_FONT", 1, 215/255, 0)
local GREEN_FONT = CreateColorFont("GREEN_FONT", 0, 1, 0)
local LIGHT_GREEN_FONT = CreateColorFont("LIGHT_GREEN_FONT", 0.25, 0.75, 0.25)
local ORANGE_FONT = CreateColorFont("ORANGE_FONT", 1, 0.5, 0.25)
local RED_FONT = CreateColorFont("RED_FONT", 0.9, 0, 0)
local YELLOW_FONT = CreateColorFont("YELLOW_FONT", 1, 1, 0)
local WHITE_FONT = CreateColorFont("WHITE_FONT", 1, 1, 1)

local HEADER_FONT = CreateColorFont("HEADER_FONT", 1, 1, 1, _G.GameTooltipHeaderText) -- white
local INSTRUCTION_FONT = CreateColorFont("INSTRUCTION_FONT", 0, 1, 0) -- Green
INSTRUCTION_FONT:SetFont(		-- Make it smaller
	select(1,INSTRUCTION_FONT:GetFont()),
	tonumber(select(2,INSTRUCTION_FONT:GetFont())-2)
)

-- Font for reputation standings
local function ColorGradient(perc, ...)
	-- from http://www.wowpedia.org/ColorGradient
	if perc >= 1 then
		local r, g, b = select(select('#', ...) - 2, ...)
		return r, g, b
	elseif perc <= 0 then
		local r, g, b = ...
		return r, g, b
	end

	local num = select('#', ...) / 3

	local segment, relperc = _G.math.modf(perc*(num-1))
	local r1, g1, b1, r2, g2, b2 = select((segment*3)+1, ...)

	return r1 + (r2-r1)*relperc, g1 + (g2-g1)*relperc, b1 + (b2-b1)*relperc
end

local STANDING_FONT_8 = CreateColorFont("STANDING_FONT_8", ColorGradient( 1.0, 0.9,0,0, 1,1,0, 0,1,0 ))
local STANDING_FONT_7 = CreateColorFont("STANDING_FONT_7", ColorGradient( 0.8, 0.9,0,0, 1,1,0, 0,1,0 ))
local STANDING_FONT_6 = CreateColorFont("STANDING_FONT_6", ColorGradient( 0.6, 0.9,0,0, 1,1,0, 0,1,0 ))
local STANDING_FONT_5 = CreateColorFont("STANDING_FONT_5", ColorGradient( 0.4, 0.9,0,0, 1,1,0, 0,1,0 ))
local STANDING_FONT_4 = CreateColorFont("STANDING_FONT_4", ColorGradient( 0.2, 0.9,0,0, 1,1,0, 0,1,0 ))
local STANDING_FONT_3 = CreateColorFont("STANDING_FONT_3", ColorGradient( 0.0, 0.9,0,0, 1,1,0, 0,1,0 ))


-------------------------------------------------------------------------------
-- Variables
-------------------------------------------------------------------------------

local update_queued = false
local in_combat = false

local db
--local faction_db
local known_db
local trainers_db
local skills_db

------------------------------------------------------------------------------------------
-- Profession localisation and reverse localization for the skill_db
------------------------------------------------------------------------------------------

local GetStaticProfName
local GetLocalProfName
do  -- Closure

	local PROFFESSION_LOCAL = {
		ALCHEMY			= _G.GetSpellInfo(51304),
		ARCHAEOLOGY		= _G.PROFESSIONS_ARCHAEOLOGY,
		BLACKSMITHING	= _G.GetSpellInfo(51300),
		COOKING			= _G.PROFESSIONS_COOKING,
		ENCHANTING		= _G.GetSpellInfo(51313),
		ENGINEERING		= _G.GetSpellInfo(51306),
		FIRSTAID			= _G.PROFESSIONS_FIRST_AID,
		FISHING			= _G.PROFESSIONS_FISHING,
		HERBALISM		= _G.GetSpellInfo(13614), -- 9134
		INSCRIPTION		= _G.GetSpellInfo(45363),
		JEWELCRAFTING	= _G.GetSpellInfo(51311),
		LEATHERWORKING	= _G.GetSpellInfo(51302),
		MINING			= _G.GetSpellInfo(32606),
		SKINNING			= _G.GetSpellInfo(8613),
		TAILORING		= _G.GetSpellInfo(51309),
	}
	
	if IS_MOP then
		-- New cooking profession
		PROFFESSION_LOCAL.WAYOFTHBREW			= _G.GetSpellInfo(125589)
		PROFFESSION_LOCAL.WAYOFTHEOVEN		= _G.GetSpellInfo(125588)
		PROFFESSION_LOCAL.WAYOFTHEWOK			= _G.GetSpellInfo(125584)
		PROFFESSION_LOCAL.WAYOFTHGRIL			= _G.GetSpellInfo(124694)
		PROFFESSION_LOCAL.WAYOFTHPOT			= _G.GetSpellInfo(125586)
		PROFFESSION_LOCAL.WAYOFTHSTEAMER		= _G.GetSpellInfo(125587)
	end

	local PROFFESSION_REVLOCAL = {}
	for k,v in pairs(PROFFESSION_LOCAL) do
		PROFFESSION_REVLOCAL[v] = k
	end

	function GetStaticProfName(local_name, hide_error)
		local static_name =  PROFFESSION_REVLOCAL[local_name]

		if static_name then
			return static_name
		else
			if not hide_error then
				-- Give the error only one time
				-- hide_error is a special case for the dump function so that
				-- the dump function doesn't end up poluting the table with
				-- class names
				PROFFESSION_LOCAL[local_name] = local_name
				PROFFESSION_REVLOCAL[local_name] = local_name
				err("Missing profession mapping for %s", local_name)
			end
			return local_name
		end
	end

	function GetLocalProfName(static_name)
		local local_name = PROFFESSION_LOCAL[static_name]

		if local_name then
			return local_name
		else
			-- Give the error only one time
			PROFFESSION_LOCAL[static_name] = static_name
			PROFFESSION_REVLOCAL[static_name] = static_name
			err("Missing profession mapping for %s", static_name)
			return static_name
		end
	end

end -- Closure

------------------------------------------------------------------------------------------
-- Access to known_db
------------------------------------------------------------------------------------------

-- Validate if the toon knowns the skill
local function IsKnownSkill(spell_id)
	if spell_id and known_db[spell_id] then
		return true
	else
		return nil
	end
end

-- Validate if the toon knowns an ability (from GetTrainerServiceAbilityReq())
local function IsKnownAbility(ability)
	if type(ability) == "string" then
		-- The ability is known by name
		return IsKnownSkill(ABILITY_NAME_TO_SPELLID[ability])
	else
		-- The ability is known by spell_id
		return IsKnownSkill(ability)
	end
end

-- Set Known skill. Also take care of implied skills
local function SetKnownSkill(spell_id, spell_name, value)
	if not spell_id then return end
	
	--if spell_id == 115913 then err("spell_id = %s, spell_name = %s, value = %s",spell_id, spell_name, value) end
	
	value = value == nil and true or value

	--known_db[spell_id] = spell_name or true
	known_db[spell_id] = value or nil

	if value and IMPLIED_SKILLS[spell_id] then
		for _, implied_spell_id in ipairs(IMPLIED_SKILLS[spell_id]) do
			known_db[implied_spell_id] = true
		end
	end
end


------------------------------------------------------------------------------------------
-- Reputation related functions
------------------------------------------------------------------------------------------

-- Find the base cost from the real cost and the reputation standing
local function GetBaseCost(cost, StandingId)
	assert(StandingId and StandingId > 3,"Invalid reputation StandingId (must be between 4 and 8): "..tostring(StandingId))

	-- Goblin have a special discount with every faction
	-- 69044 = Best Deals Anywhere, the Goblin racial that make this happen
	if IsKnownSkill(69044) then
		StandingId = 8
	end

	local rebate = 1 - (StandingId - 4) * .05
	return floor(cost / rebate + 0.0001) -- The 0.0001 is for a strange bug that has floor(700 / .7) == 999
end

local function GetRealCost(cost, StandingId)
	assert(StandingId and StandingId > 3,"Invalid reputation StandingId (must be between 4 and 8): "..tostring(StandingId))

	-- Goblin have a special discount with every faction
	-- 69044 = Best Deals Anywhere, the Goblin racial that make this happen
	if IsKnownSkill(69044) then
		StandingId = 8
	end

	local rebate = 1 - (StandingId - 4) * .05
	return ceil(cost * rebate - 0.0001)
end

local GetReputationStanding
local GetCheapestSkillCost
do -- Closure for the local tables
	local rep_index_cache = {}

	function GetReputationStanding(rep_name)

		if rep_index_cache[rep_name] then
			local name, _, standingId = _G.GetFactionInfo(rep_index_cache[rep_name])

			if name == rep_name then
				return standingId
			else
				-- Reputations have change, the index is no longer valid
				rep_index_cache[rep_name] = nil
			end
		end

		for factionIndex = 1, _G.GetNumFactions() do
			local name, _, standingId, _, _, _, _, _, isHeader = _G.GetFactionInfo(factionIndex)
			if not isHeader and rep_name == name then
				rep_index_cache[rep_name] = factionIndex
				return standingId
			end
		end

		return 0
	end

	local rep_standings  = {}
	local best_rep = {}

	function GetCheapestSkillCost(skill)
		local cost = skill.cost
		local rep_table = skill.valid_rep

		if not rep_table then
			return cost, ""
		end
		local best_standing = 0

		-- Refesh the rep_standings
		for factionIndex = 1, _G.GetNumFactions() do
			local rep_name, _, standingId, _, _, _, _, _, isHeader = _G.GetFactionInfo(factionIndex)

			if not isHeader then
				-- We refresh the index cache while were's at it
				rep_index_cache[rep_name] = factionIndex
				rep_standings[rep_name] = standingId
			end
		end

		-- Find the best standing
		for rep_name in pairs(rep_table) do
			if rep_standings[rep_name] and rep_standings[rep_name] > best_standing then
				best_standing = rep_standings[rep_name]
			end
		end

		if best_standing == 0 then return cost, "" end

		-- Build a string of the best Reputation factions
		wipe(best_rep)

		for rep_name in pairs(rep_table) do
			if rep_standings[rep_name] == best_standing then
				tinsert(best_rep, rep_name)
			end
		end
		sort(best_rep)

		return GetRealCost(cost, best_standing), concat(best_rep, ", ")
	end

end -- End of closure


------------------------------------------------------------------------------------------
-- Scaning functions
------------------------------------------------------------------------------------------

-- Scan the current class spells to see what the toon already know
function Syllabus:InitKnownspells()
	-- Find ability names
	for spell_id in pairs(IS_ABILITY) do
		local name, rank = _G.GetSpellInfo(spell_id)
		-- In English, the rank is put between () in the requires text from the skills,
		-- I hope it's the same in the other locales
		local full_name = (rank and rank ~= "" and ("%s (%s)"):format(name, rank)) or name
		ABILITY_NAME[spell_id] = full_name
		SHORT_ABILITY_NAME[spell_id] = name
		ABILITY_NAME_TO_SPELLID[full_name] = spell_id
	end

	for spell_id in pairs(IS_ABILITY_EXCEPTION) do
		local name, rank = _G.GetSpellInfo(spell_id)
		local full_name = (rank and rank ~= "" and ("%s (%s)"):format(name, rank)) or name
		IS_ABILITY_EXCEPTION_NAME[full_name] = spell_id
	end

	-- Find the highest index value for the spell books
	local max_spell_index
	for i = 1, 20 do
	   local name, _, offset, numSlots = _G.GetSpellTabInfo(i)
	   if name then 
	      max_spell_index = offset + numSlots 
	   else
	      break
	   end
	end

	-- Scan the Spellbook
	for i = 1, max_spell_index do
		local skill_type, spell_id = _G.GetSpellBookItemInfo(i, _G.BOOKTYPE_SPELL)
		if skill_type then
			local skill_name = (spell_id and _G.GetSpellInfo(spell_id)) or nil
			if skill_name then
				SetKnownSkill(spell_id, skill_name, _G.IsPlayerSpell(spell_id))
			end
		end
	end
	
	-- Scan the mount
	for i = 1, _G.GetNumCompanions("MOUNT") do
		local spell_id = select(3, _G.GetCompanionInfo("MOUNT", i))
		SetKnownSkill(spell_id, "MOUNT")
	end

	-- Scan the trackings
	if TRACKING[MY_CLASS_UPPER] then
		for i = 1, _G.GetNumTrackingTypes() do
			-- Yes, the only way to find the spell without knowing the name is by the texture...
			local name, texture = _G.GetTrackingInfo(i)
			if TRACKING[MY_CLASS_UPPER][texture] then
				SetKnownSkill(TRACKING[MY_CLASS_UPPER][texture], name)
			end
		end
	end

	-- Copy the shared skills to the class skills
	--CopyTable(faction_db.shared, faction_db.classes[MY_CLASS_UPPER])
	CopyTable(skills_db.all_classes, skills_db.classes[MY_CLASS_UPPER])

	-- Racial exceptions
	-- If only a certain race can train a skill and the toon is not of that
	-- race, the skill as known. It's a shortcut that we need to remember.
	-- Not sure if this is a problem with race change
	for spell_id, races in pairs(SKILL_RACIAL_EXCEPTIONS) do
		if not _G.tContains(races, MY_RACE_EN) then
			SetKnownSkill(spell_id,"SKILL_RACIAL_EXCEPTIONS")
		end
	end
end

local function ScanProfessionInfo(index, show_option, prof_seen)
	if not index then return end

	local prof_name, _, rank, max_rank = _G.GetProfessionInfo(index)

	-- Update to current values
	MY_PROFESSIONS[prof_name] = MY_PROFESSIONS[prof_name] or {}
	MY_PROFESSIONS[prof_name].rank = rank
	MY_PROFESSIONS[prof_name].max_rank = max_rank
	MY_PROFESSIONS[prof_name].show_option = show_option

	if not TITLES[prof_name] then return end

	local prof_ranks = _G.PROFESSION_RANKS

	-- We have to infer some training from the max rank
	for i = 1, #prof_ranks do
		local value = prof_ranks[i][1]

		if max_rank >= value then
			-- We know that one
			SetKnownSkill(TITLES[prof_name][value], _G.GetSpellInfo(TITLES[prof_name][value]))
		else
			break
		end
	end
end

local function ScanUnspentTalentPoints()
	local primary, secondary

	primary = GetUnspentTalentPoints(false, false, 1)

	if GetNumTalentGroups() > 1 then
		-- Secondary talent has been bought
		secondary = GetUnspentTalentPoints(false, false, 2)
	end

	-- Secondary will be nil if not available, 0 if available but no unspent point
	return primary, secondary

end

------------------------------------------------------------------------------------------
-- Data broker text
------------------------------------------------------------------------------------------

-- Generate the text that will be used for the data broker bar
local function GetUpdatedBrokerText()
	-- Find the professions that need to be trained
	local total_learnable, total_cost = 0, 0
	local cost_type = db.profile.broker_cost

	for my_prof_name, my_prof in pairs(MY_PROFESSIONS) do
		if db.profile[MY_PROFESSIONS[my_prof_name].show_option] then
			for prof_name, prof in pairs(skills_db.professions) do
				if prof_name == my_prof_name then
					for spell_id, skill in pairs(prof) do
						--err("pn = %s, sn = %s, mpn = %s",prof_name,skill_name,my_prof_name)
						if not IsKnownSkill(spell_id) and
							skill[MY_FACTION_FLAG] and
							(not skill.rank_req or skill.rank_req <= my_prof.rank) and
							(not skill.level_req or skill.level_req <= MY_LEVEL) and
							(not skill.requires or IsKnownAbility(skill.requires)) and
							(not NEED_ARMOR[spell_id]
							 or NEED_ARMOR[spell_id] == ARMOR_BY_CLASS[MY_CLASS_UPPER]
							 or NEED_ARMOR[spell_id] == MY_CLASS_UPPER) and
							(db.profile.broker_filter_cost == 0 or skill.cost <= db.profile.broker_filter_cost * 10000)
						then
							total_learnable = total_learnable + 1

							if cost_type == 'base' then
								total_cost = total_cost + skill.cost
							elseif cost_type == 'cheapest' then
								total_cost = total_cost + GetCheapestSkillCost(skill)
							end
						end
					end
				end
			end
		end
	end

	-- Find the class skills that need to be trained
	if skills_db.classes[MY_CLASS_UPPER] and db.profile.show_class then
		-- When the database is empty because no trainer have been seen yet
		for spell_id, skill in pairs(skills_db.classes[MY_CLASS_UPPER]) do
			if not IsKnownSkill(spell_id) and
				skill[MY_FACTION_FLAG] and
				(not skill.level_req or skill.level_req <= MY_LEVEL)  and
				(not skill.requires or IsKnownAbility(skill.requires))  and
				(db.profile.broker_filter_cost == 0 or skill.cost <= db.profile.broker_filter_cost * 10000) and
				-- Filter out riding skills if not wanted
				(db.profile.show_riding or (not RIDING_SKILLS[skill.spell_id]))	then
				total_learnable = total_learnable + 1

				if cost_type == 'base' then
					total_cost = total_cost + skill.cost
				elseif cost_type == 'cheapest' then
					total_cost = total_cost + GetCheapestSkillCost(skill)
				end
			end
		end
	end

	if db.profile.show_unspent_talents ~= 'never' then
		-- Add one learnable for primary and one for secondary talents even if multiple points are missing
		local primary, secondary = ScanUnspentTalentPoints()

		total_learnable = total_learnable + (primary > 0 and 1 or 0) + (secondary and secondary > 0 and 1 or 0)
	end

	local text = ""

	if db.profile.broker_learnable then
		text = (total_learnable ~= 0) and total_learnable or db.profile.broker_nothing_msg
	end

	if cost_type ~= 'none' and total_cost ~= 0 then
		if text ~= "" then
			text = text .. ": "
		end
		text = text .. _G.GetMoneyString(total_cost)
	end

	return text
end

local prof_seen = {}
function Syllabus:UpdateSkills()
	if in_combat then
		update_queued = true
		return
	end

	-- Professions related stuff
	--wipe(MY_PROFESSIONS)
	-- seen is used to erase the profession that have been unlearned without wipings the whole table.
	wipe(prof_seen)

	local prof1, prof2, arch, fish, cook, firstaid = _G.GetProfessions()
	--prof_seen[prof1] = true
	--prof_seen[prof2] = true
	--prof_seen[arch] = true
	--prof_seen[fish] = true
	--prof_seen[cook] = true
	--prof_seen[firstaid] = true

	ScanProfessionInfo(prof1, 'show_prof1', prof_seen)
	ScanProfessionInfo(prof2, 'show_prof2', prof_seen)
	ScanProfessionInfo(arch, 'show_arch', prof_seen)
	ScanProfessionInfo(fish, 'show_fish', prof_seen)
	ScanProfessionInfo(cook, 'show_cook', prof_seen)
	ScanProfessionInfo(firstaid, 'show_firstaid')
	
	-- Scanning for Mists cooking sub-professions and other such exceptions
	for _, spell_id in ipairs(EXTRA_PROFESSIONS) do
		local prof_name = _G.GetSpellInfo(spell_id)
		if _G.IsPlayerSpell(spell_id) and not IsKnownSkill(spell_id) then
			prof_seen[prof_name] = true

			-- Need away to find the rank
			-- Update to current values (hardcoded minimum and maximum values for unknown)
			MY_PROFESSIONS[prof_name] 					= MY_PROFESSIONS[prof_name] or {}
			MY_PROFESSIONS[prof_name].rank 			= MY_PROFESSIONS[prof_name].rank or 530
			MY_PROFESSIONS[prof_name].max_rank 		= MY_PROFESSIONS[prof_name].max_rank or 600
			MY_PROFESSIONS[prof_name].show_option	= 'show_cook'
		end
		
		-- We set the value even if false in case the skill was unlearned
		SetKnownSkill(spell_id, prof_name, _G.IsPlayerSpell(spell_id))
	end

	-- Class related stuff
	MY_LEVEL = _G.UnitLevel("player")
	
	-- Erase any proficiency that was droped
	--for k in pairs(MY_PROFESSIONS) do
	--	if not prof_seen[k] then MY_PROFESSIONS[k] = nil end
	--end

	-- Update the broker text
	DataObj.text = GetUpdatedBrokerText()
end


------------------------------------------------------------------------------------------
-- Tooltip Autohide
------------------------------------------------------------------------------------------

local SetMainAutohide
local SetSkillTooltip
local GetSkillTooltip
local ReleaseSkillTooltip
do
	local main_tooltip
	local skill_tooltip
	local tooltip_anchor
	local elapsed = 0

	local OnUpdate_frame = _G.CreateFrame("frame")
	OnUpdate_frame:Hide() -- to prevent the OnUpdate until it is needed.
	OnUpdate_frame:SetScript("OnUpdate",
				 function(self, elap)
					 elapsed = elapsed + elap

					 if (main_tooltip and main_tooltip:IsMouseOver()) or
					    (tooltip_anchor and tooltip_anchor:IsMouseOver()) or
					    (skill_tooltip and skill_tooltip:IsMouseOver()) then
					 	 elapsed = 0
					 	 return
					 end

					 if elapsed < .35 then
						 return
					 end

					 skill_tooltip = skill_tooltip and QTip:Release(skill_tooltip)
					 main_tooltip = main_tooltip and QTip:Release(main_tooltip)
					 tooltip_anchor = nil
					 OnUpdate_frame:Hide()
				 end)

	-- Set the Autohide for the main tooltip
	function SetMainAutohide(tooltip, anchor)
		main_tooltip = tooltip
		tooltip_anchor = anchor
		elapsed = 0
		OnUpdate_frame:Show() -- Start the timer
	end

	-- Add an extra tooltip for the autohide "cluster"
	function SetSkillTooltip(tooltip)
		skill_tooltip = tooltip
		elapsed = 0
	end

	-- Retreive the extra tooltip
	function GetSkillTooltip()
		return skill_tooltip
	end

	-- Manually release the extra tooltip
	function ReleaseSkillTooltip()
		skill_tooltip = skill_tooltip and QTip:Release(skill_tooltip)
		return skill_tooltip
	end
end

------------------------------------------------------------------------------------------
-- Line Tooltips
------------------------------------------------------------------------------------------

local SetInfoTooltipType
local ToggleInfoTooltip
local ShowInfoTooltip
local HideInfoTooltip
local ShowTextTooltip
local HideTextTooltip
do
	local function SetTomTomArrow(frame, trainer_index, button)
		local prof_name, trainer_id = trainer_index:match("(.+)[$][$](.+)")
		local trainer = trainers_db[tonumber(trainer_id)]
		prof_name = prof_name == MY_CLASS_UPPER and MY_CLASS or prof_name

		-- Track an error reported by Zidomo
		assert(trainer,("trainer was nil: trainer_index = %s, prof_name = %s, trainer_id = %s"):format(trainer_index or 'nil',prof_name or 'nil',trainer_id or 'nil'))

		-- We use map_id
		if trainer.map_id then
			_G.TomTom:AddMFWaypoint(
				trainer.map_id,
				trainer.floor or 0,
				floor(trainer.x*1000 + .5)/1000,
				floor(trainer.y*1000 + .5)/1000,
				{
            	title = ("%s (%s)"):format(trainer.name or trainer_id, prof_name),
            	persistent = false,
        		}
			)
		elseif trainer.zone_id and trainer.continent_id then
			_G.TomTom:AddZWaypoint(
				trainer.continent_id,
				trainer.zone_id,
				floor(trainer.x*1000 + .5)/10,
				floor(trainer.y*1000 + .5)/10,
				("%s (%s)"):format(trainer.name or trainer_id, prof_name),
				false
			)
		end
	end

	local spell_tip = _G.CreateFrame("GameTooltip", "Syllabus_SpellTooltip", _G.UIParent, "GameTooltipTemplate")

	local function ShowSkillTooltip(frame, spell_id)
		if spell_id then
			spell_tip:SetOwner(frame, "ANCHOR_NONE")
			spell_tip:ClearAllPoints()
			--Add TipTac Support
			--if _G.TipTac and _G.TipTac.AddModifiedTip and not spell_tip.tiptac then
			--	_G.TipTac:AddModifiedTip(spell_tip)
			--	spell_tip.tiptac = true
			--end

			spell_tip:SetScale(db.profile.tooltip_scale)
			spell_tip:SetClampedToScreen(true)
			spell_tip:SetSpellByID(spell_id)

			if (frame:GetLeft() + 15) < spell_tip:GetWidth() then
				spell_tip:SetPoint("TOPLEFT", frame, "TOPRIGHT", 10, 6)
			else
				spell_tip:SetPoint("TOPRIGHT", frame, "TOPLEFT", -10, 6)
			end
			spell_tip:Show()
		end
	end

	local function HideSkillTooltip()
		spell_tip:Hide()
	end

	local function HideTrainerTooltip()
		ReleaseSkillTooltip()
	end

	local standing_fonts = {
		[8] = STANDING_FONT_8,
		[7] = STANDING_FONT_7,
		[6] = STANDING_FONT_6,
		[5] = STANDING_FONT_5,
		[4] = STANDING_FONT_4,
		[3] = STANDING_FONT_3,
		[2] = STANDING_FONT_3,
		[1] = STANDING_FONT_3,
		[0] = STANDING_FONT_4,
	}
	local trainer_standing = {}
	local trainer_sorted = {}

	-- Sort trainers based on standing, zone, or name.
	--local sort_trainers_ref		-- Only to be used for sort
	local function SortTrainers(a, b)
		local ta, tb = trainers_db[a], trainers_db[b]
		if trainer_standing[a] ~= trainer_standing[b] then
			return trainer_standing[a] > trainer_standing[b]
		elseif ta.zone_id ~= tb.zone_id or ta.continent_id ~= tb.continent_id then
		   local zone_a = ta.continent_id and ZONE_NAMES[ta.continent_id][ta.zone_id] or ""
		   local zone_b = tb.continent_id and ZONE_NAMES[tb.continent_id][tb.zone_id] or ""
		   --err("a = %s, zone_a = %s [%s, %s], b = %s, zone_b = %s [%s, %s], result = %s",a,zone_a,ta.continent_id,ta.zoneid,b,zone_b,tb.continent_id,tb.zoneid,zone_a < zone_b)
			return zone_a < zone_b
		else
			return (ta.name or a) < (tb.name or b)
		end
	end

	local function ShowTrainerTooltip(frame, spell_id, skill, prof_name, prof_type)
		local line, column, tomtom_on
		local skill_name = skill.name or _G.GetSpellInfo(spell_id)

		wipe(trainer_standing)

		for trainer_id in pairs(skill.trainers) do
			trainer_standing[trainer_id] = GetReputationStanding(trainers_db[trainer_id].rep)
		end

		wipe(trainer_sorted)

		for trainer_id in pairs(skill.trainers) do
			-- Only use the trainer avialable for the faction
			if trainers_db[trainer_id][MY_FACTION_FLAG] then
				tinsert(trainer_sorted, trainer_id)
			end
		end

		--sort_trainers_ref = trainers
		sort(trainer_sorted, SortTrainers)

		local trainer_tt = GetSkillTooltip() or QTip:Acquire(MODNAME.."TrainerTooltip", 3, "LEFT", "CENTER", "RIGHT")
		SetSkillTooltip(trainer_tt)
		trainer_tt:SetScale(db.profile.tooltip_scale)
		trainer_tt:SetHeaderFont(HEADER_FONT)
		trainer_tt:SetFont(GOLD_FONT)
		trainer_tt:Clear()

		line, column = trainer_tt:AddHeader()
		trainer_tt:SetCell(line, 1, ("%s %s"):format(L["Trainers for"], skill_name), HEADER_FONT, "CENTER", 3)
		trainer_tt:AddSeparator()

		if #trainer_sorted == 0 then
			-- no trainer information
			line, column = trainer_tt:AddLine()
			trainer_tt:SetCell(line, 1, L["No trainer available for this skill."], "CENTER")
		else
			local first_line = true
			local standing, reputation
			for i = 1,#trainer_sorted do
				local trainer_id = trainer_sorted[i]
				if standing ~= trainer_standing[trainer_id] then
					-- New reputation paragraph
					standing = trainer_standing[trainer_id]
					reputation = _G["FACTION_STANDING_LABEL"..standing] or _G.UNKNOWN
					if not first_line then
--						trainer_tt:AddSeparator(4,0,0,0,0) -- Empty line
						trainer_tt:AddSeparator(nil,0,0,0,0) -- Empty line
						trainer_tt:AddSeparator(nil,0,0,0,0) -- Empty line
--						line, column = trainer_tt:AddLine()
--						trainer_tt:SetCell(line, 1, " ", "CENTER")
						first_line = false
					end

					line, column = trainer_tt:AddHeader()
					if standing >= 4 or standing == 0 then
						trainer_tt:SetCell(
							line,
							1,
							("%s (%s)"):format(
								reputation,
								_G.GetMoneyString(GetRealCost(skill.cost, standing >= 3 and standing or 4))
							),
							HEADER_FONT,
							"LEFT",
							3
						)
					else
						trainer_tt:SetCell(line, 1, reputation, HEADER_FONT, "LEFT", 3)
					end
				end

				local trainer = trainers_db[trainer_id]
				line, column = trainer_tt:AddLine()
				trainer_tt:SetCell(line, 1, "  "..(trainer.name or trainer_id), standing_fonts[standing])
				local zone_name = MAPID_TO_NAME[trainer.map_id] or ZONE_NAMES[trainer.continent_id][trainer.zone_id]
				if trainer.x and trainer.y then
					trainer_tt:SetCell(
						line,
						3,
						("%s (%.1f, %.1f)"):format(zone_name, trainer.x*100, trainer.y*100),
						standing_fonts[standing]
					)
				else
					trainer_tt:SetCell(line, 3, zone_name, standing_fonts[standing])
				end

				-- TomTom integration
				if _G.TomTom and trainer.continent_id and trainer.continent_id ~= -1 then
					-- If continent_id is there, zone_id, x and y are there
					-- continent_id == -1 is the Cosmic zone and cannot be mapped by TomTom
					local trainer_index = prof_name .. "$$" .. trainer_id
					trainer_tt:SetLineScript(line, "OnMouseUp", SetTomTomArrow, trainer_index)
					tomtom_on = true -- TomTom hook was set on at least one line
				end

			end

			-- Instruction on how to get the TomTom arrow
			if tomtom_on then
				trainer_tt:AddSeparator()
				line, column = trainer_tt:AddLine()
				trainer_tt:SetCell(line, 1, L['Click a trainer line to set the Tomtom arrow'], INSTRUCTION_FONT, "CENTER", 3)
			end
		end

		--trainer_tt:UpdateScrolling()
		if (frame:GetLeft() + 15) < (trainer_tt:GetWidth()) then
			trainer_tt:SetPoint("TOPLEFT", frame, "TOPRIGHT", 10, 6)
		else
			trainer_tt:SetPoint("TOPRIGHT", frame, "TOPLEFT", -10, 6)
		end
		trainer_tt:Show()

	end

	local current_tooltip = 'skill'
	function SetInfoTooltipType(tooltip_type)
		current_tooltip = tooltip_type
	end

	function ShowInfoTooltip(frame, skill_index, button)
		-- Note: I didn't find a way to get userdata in the line frame and I didn't want to pass
		-- a table that would generate garbage. If you have a better idea, let me know but in any
		-- cases, I want to change the database format so use the spell_id as key and I think I'll be
		-- able to just pass a reference to the skill table at that point.
		local prof_type, prof_name, spell_id = skill_index:match("(.+)[$][$](.+)[$][$](.+)")
		spell_id = tonumber(spell_id)
		if current_tooltip == 'skill' then
			ShowSkillTooltip(frame, spell_id)
		else
			local skill = skills_db[prof_type][prof_name][spell_id]
			ShowTrainerTooltip(frame, spell_id, skill, prof_name, prof_type)
		end
	end

	function HideInfoTooltip(frame, force_hide)
		--err("force_hide = %s", tostring(force_hide))
		if current_tooltip == 'skill' then
			HideSkillTooltip()
		else
			-- The Trainer tooltip auto-hides, it needs to be hiden
			-- only when toggling from one tooltip to another.
			if force_hide then HideTrainerTooltip() end
		end
	end

	function ToggleInfoTooltip(frame, skill_index, button)
		HideInfoTooltip(frame, true)
		current_tooltip = current_tooltip == 'trainer' and 'skill' or 'trainer'
		ShowInfoTooltip(frame, skill_index, button)
	end

	function ShowTextTooltip(frame, text, button)
--err("ShowTextTooltip: text=%s, frame = %s", text, frame)
		if text then
			-- Hide the skill tooptip if present
			HideInfoTooltip(frame, true)

			spell_tip:SetOwner(frame, "ANCHOR_NONE")
			spell_tip:ClearAllPoints()

			spell_tip:SetScale(db.profile.tooltip_scale)
			spell_tip:SetClampedToScreen(true)
			spell_tip:AddLine(text)

			if (frame:GetLeft() + 15) < spell_tip:GetWidth() then
				spell_tip:SetPoint("TOPLEFT", frame, "TOPRIGHT", 10, 6)
			else
				spell_tip:SetPoint("TOPRIGHT", frame, "TOPLEFT", -10, 6)
			end
			spell_tip:Show()
		end
	end

	function HideTextTooltip()
--err("HideTextTooltip")
		spell_tip:Hide()
	end
end -- InfoTooltip closure

------------------------------------------------------------------------------------------
-- Main Tooltips
------------------------------------------------------------------------------------------

-- Sort functions
local sort_prof_ref
local function SortRankReq(a,b)
	local name_a = sort_prof_ref[a].name or _G.GetSpellInfo(a) or ""
	local name_b = sort_prof_ref[b].name or _G.GetSpellInfo(b) or ""
	if sort_prof_ref[a].rank_req ~= sort_prof_ref[b].rank_req then
		return (sort_prof_ref[a].rank_req or 0) < (sort_prof_ref[b].rank_req or 0)
	elseif name_a ~= name_b then
		return name_a < name_b
	else
		return a < b
	end
end

local sort_class_to_display_ref
local function SortClassLevel(a,b)
	local name_a = sort_class_to_display_ref[a].name or _G.GetSpellInfo(a) or ""
	local name_b = sort_class_to_display_ref[b].name or _G.GetSpellInfo(b) or ""
	if sort_class_to_display_ref[a].level_req ~= sort_class_to_display_ref[b].level_req then
		return (sort_class_to_display_ref[a].level_req or 0) < (sort_class_to_display_ref[b].level_req or 0)
	elseif name_a ~= name_b then
		return name_a < name_b
	else
		return a < b
	end
end

local function ShowTooltip(anchor)
	local total_learnable, total_cost = 0, 0
	local cost_type = db.profile.display_cost

	------------------------------------------------------------------------------------------
	-- Find the professions that need to be trained
	------------------------------------------------------------------------------------------
	local prof_to_display = AcquireTable()

	for my_prof_name, my_prof in pairs(MY_PROFESSIONS) do
		if db.profile[MY_PROFESSIONS[my_prof_name].show_option] then
			prof_to_display[my_prof_name] = AcquireTable()

			for prof_name, prof in pairs(skills_db.professions) do
				if prof_name == my_prof_name then
					for spell_id, skill in pairs(prof) do
						--err("pn = %s, sn = %s, mpn = %s",prof_name,skill_name,my_prof_name)
						if not IsKnownSkill(spell_id) and
							skill[MY_FACTION_FLAG] and
							(not skill.rank_req or skill.rank_req <= my_prof.rank + db.profile.skill_range) and
							(not NEED_ARMOR[spell_id] or
							 NEED_ARMOR[spell_id] == ARMOR_BY_CLASS[MY_CLASS_UPPER] or
							 NEED_ARMOR[spell_id] == MY_CLASS) and
							 -- If the aiblity is an exception, the line is not displayed unless
							 -- the toon knowns that ability
							(not skill.requires
							 or not (IS_ABILITY_EXCEPTION[skill.requires] or IS_ABILITY_EXCEPTION_NAME[skill.requires])
							 or IsKnownAbility(skill.requires)) and
							(db.profile.filter_cost == 0 or skill.cost <= db.profile.filter_cost * 10000)
						then
							--err("Adding %s:%s %s",tostring(my_prof_name),tostring(skill_name),tostring(skill.rank_req))
							local display_skill = AcquireTable()
							display_skill.rank_req = skill.rank_req
							display_skill.level_req = skill.level_req
							display_skill.requires = skill.requires
							display_skill.icon = skill.icon or _G.GetSpellTexture(spell_id)
--if spell_id == 110412 then err("display_skill.icon = %s, skill.icon = %s",display_skill.icon,skill.icon) end
							--display_skill.item_link = skill.item_link
							--display_skill.spell_id = skill.spell_id
							display_skill.name = skill.name
							display_skill.cost, display_skill.best_rep = GetCheapestSkillCost(skill)

							if cost_type == 'base' then
								-- We reset to base cost if require
								display_skill.cost = skill.cost
							elseif cost_type == 'none' then
								-- We blank the cost if require
								display_skill.cost = nil
							end

							-- Back reference to give the tooltip
							display_skill.orig = skill

							prof_to_display[my_prof_name][spell_id] = display_skill

							-- Running totals
							if (not skill.rank_req or skill.rank_req <= my_prof.rank) and
								(not skill.level_req or skill.level_req <= MY_LEVEL) and
								(not skill.requires or IsKnownAbility(skill.requires)) then
								total_learnable = total_learnable + 1
								total_cost = total_cost + (display_skill.cost or 0)
							end
						end
					end
				end
			end
		end
	end

	------------------------------------------------------------------------------------------
	-- Find the class skills that need to be trained
	------------------------------------------------------------------------------------------
	local class_to_display = AcquireTable()

	if skills_db.classes[MY_CLASS_UPPER] and db.profile.show_class then
		-- When the database is empty because no trainer have been seen yet
		for spell_id, skill in pairs(skills_db.classes[MY_CLASS_UPPER]) do
			if not IsKnownSkill(spell_id) and
				skill[MY_FACTION_FLAG] and
				(not skill.level_req or skill.level_req <= MY_LEVEL + db.profile.level_range) and
				 -- If the aiblity is an exception, the line is not displayed unless
				 -- the toon knowns that ability
				(not skill.requires or not IS_ABILITY_EXCEPTION_NAME[skill.requires]
				 or IsKnownAbility(skill.requires)) and
				(db.profile.filter_cost == 0 or skill.cost <= db.profile.filter_cost * 10000) and
				-- Filter out riding skills if not wanted
				(db.profile.show_riding or not RIDING_SKILLS[spell_id])
			then
				local display_skill = AcquireTable()

				display_skill.rank_req = skill.rank_req
				display_skill.level_req = skill.level_req
				display_skill.requires = skill.requires
				display_skill.icon = skill.icon or _G.GetSpellTexture(spell_id)
				--display_skill.item_link = skill.item_link
				display_skill.name = skill.name
				display_skill.cost, display_skill.best_rep = GetCheapestSkillCost(skill)

				if cost_type == 'base' then
					-- We reset to base cost if require
					display_skill.cost = skill.cost
				elseif cost_type == 'none' then
					-- We blank the cost if require
					display_skill.cost = nil
				end

				class_to_display[spell_id] = display_skill

				-- Running totals
				if (not skill.level_req or skill.level_req <= MY_LEVEL) and
					(not skill.requires or IsKnownAbility(skill.requires)) then
					total_learnable = total_learnable + 1
					total_cost = total_cost + (display_skill.cost or 0)
				end
			end
		end
	end

	------------------------------------------------------------------------------------------
	-- Display
	------------------------------------------------------------------------------------------
	-- Sort alpha for the professions
	local sort_prof = AcquireTable()

	for prof_name in pairs(prof_to_display) do
		tinsert(sort_prof,prof_name)
	end
	sort(sort_prof)

	local tooltip, line, column

	-- Reset the tooltip type for the lines
	SetInfoTooltipType('skill')

	tooltip = QTip:Acquire(MODNAME.."Tooltip", 3, "LEFT", "CENTER", "RIGHT")

	tooltip:Clear()
	tooltip:SetScale(db.profile.tooltip_scale)
	tooltip:SetHeaderFont(HEADER_FONT)
	tooltip:SetFont(GOLD_FONT)
	tooltip:SmartAnchorTo(anchor)
	SetMainAutohide(tooltip, anchor)

	line, column = tooltip:AddHeader()
	tooltip:SetCell(line, 1,
			("%s - %s %d %s |cff%s%s|r"):format(MY_NAME, _G.LEVEL, MY_LEVEL, MY_RACE, CLASS_COLORS[MY_CLASS_UPPER], MY_CLASS),
			HEADER_FONT, "CENTER", 3)
	tooltip:AddSeparator()

	local first_line = true

	for _, prof_name in ipairs(sort_prof) do
		local prof = prof_to_display[prof_name]
		--err("prof_name = %s - %s",prof_name, tostring(prof))

		if not first_line then
--			line, column = tooltip:AddSeparator(4,0,0,0,0)
			line, column = tooltip:AddSeparator(nil,0,0,0,0)
			line, column = tooltip:AddSeparator(nil,0,0,0,0)
--			line, column = tooltip:AddLine()
--			tooltip:SetCell(line, 1, " ", "CENTER")
		else
			first_line = nil
		end
		local cur_prof = MY_PROFESSIONS[prof_name]

		line, column = tooltip:AddHeader()
		tooltip:SetCell(line, 1, ("%s (%s/%s)"):format(prof_name, cur_prof.rank, cur_prof.max_rank), HEADER_FONT, "LEFT")

		-- Sort profession skills by rank and then alpha
		local sort_skill = AcquireTable()

		for spell_id, skill in pairs(prof) do
			tinsert(sort_skill,spell_id)
			--err("Adding for sort %s:%s",tostring(prof_name),tostring(skill_name))
		end
		sort_prof_ref = prof
		sort(sort_skill, SortRankReq)

		for _, spell_id in ipairs(sort_skill) do
			local skill = prof[spell_id]
			local skill_name = skill.name or _G.GetSpellInfo(spell_id)
			local text
			local req_text, separator = "", ""
			local can_train = true

			--err("Display %s:%s %s",tostring(prof_name),tostring(skill_name),tostring(skill.rank_req))
			if skill.rank_req then
				if skill.rank_req > cur_prof.rank then
					req_text = ("%s\124cFFFF0000%s\124r"):format(req_text, skill.rank_req)
					can_train = false
				else
					req_text = ("%s\124cFF00FF00%s\124r"):format(req_text, skill.rank_req)
				end
				separator = ", "
			end

			if skill.level_req and skill.level_req > 1 then
				if skill.level_req > MY_LEVEL then
					req_text = ("%s%s\124cFFFF0000%s %s\124r"):format(req_text, separator, _G.LEVEL, skill.level_req)
					can_train = false
				else
					req_text = ("%s%s\124cFF00FF00%s %s\124r"):format(req_text, separator, _G.LEVEL, skill.level_req)
				end
				separator = ", "
			end

			if skill.requires then
				-- Ability requirement are only displayed if missing
				if not IsKnownAbility(skill.requires)  then
					if db.profile.ability_req then
						local display = SHORT_ABILITY_NAME[skill.requires] or skill.requires
						req_text = ("%s%s\124cFFFF0000%s\124r"):format(req_text, separator, display)
					end
					can_train = false
				end
				--separator = ", "
			end

			if req_text ~= "" then
				req_text = " - "..req_text
			end

			local line_colour = can_train and "\124cFF00FF00" or ""

			if not skill.icon then 
				err("Missing icon: skill name = %s, spell_id = %s", skill_name, spell_id)
				skill.icon = [[INTERFACE\ICONS\INV_MISC_QUESTIONMARK]]
			end
			text = ("%s  \124T%s:0:0\124t%s%s"):format(line_colour, skill.icon, skill_name, req_text)

			line, column = tooltip:AddLine()
			tooltip:SetCell(line, 1, text)

			if skill.cost then
				if db.profile.reputation then
					tooltip:SetCell(line, 2, line_colour .. skill.best_rep)
				end
				tooltip:SetCell(line, 3, _G.GetMoneyString(skill.cost), WHITE_FONT)
			end

			local skill_index = 'professions$$' .. prof_name .. '$$' .. spell_id

			tooltip:SetLineScript(line, "OnEnter", ShowInfoTooltip, skill_index)
			tooltip:SetLineScript(line, "OnLeave", HideInfoTooltip)
			tooltip:SetLineScript(line, "OnMouseUp", ToggleInfoTooltip, skill_index)
		end

		ReleaseTable(sort_skill)
	end
	ReleaseTable(sort_prof)

	-- Only display class section if there is something to display	???
	if db.profile.show_class then
		-- Header for the class
		if not first_line then
--			line, column = tooltip:AddSeparator(4,0,0,0,0)
			line, column = tooltip:AddSeparator(nil,0,0,0,0)
			line, column = tooltip:AddSeparator(nil,0,0,0,0)
--			line, column = tooltip:AddLine()
--			tooltip:SetCell(line, 1, " ", "CENTER")
		else
			first_line = nil
		end
		line, column = tooltip:AddHeader()
		tooltip:SetCell(line, 1, ("%s (%s %d)"):format(MY_CLASS, _G.LEVEL, MY_LEVEL), HEADER_FONT, "LEFT")

		-- Sort class skills by level and then alpha
		local sort_class_skill = AcquireTable()
		for spell_id in pairs(class_to_display) do
			tinsert(sort_class_skill,spell_id)
		end
		sort_class_to_display_ref = class_to_display
		sort(sort_class_skill, SortClassLevel)

		for _, spell_id in ipairs(sort_class_skill) do
			local skill = class_to_display[spell_id]
			local skill_name = skill.name or _G.GetSpellInfo(spell_id)
			local text
			local req_text, separator = "", ""
			local can_train = true

			--err("Display %s:%s %s",tostring(prof_name),tostring(skill_name),tostring(skill.rank_req))
			if skill.level_req then
				if skill.level_req > MY_LEVEL then
					req_text = ("%s\124cFFFF0000%s %s\124r"):format(req_text, _G.LEVEL, skill.level_req)
					can_train = false
				else
					req_text = ("%s\124cFF00FF00%s %s\124r"):format(req_text, _G.LEVEL, skill.level_req)
				end
				separator = ", "
			end

			if skill.requires then
				-- Ability requirement are only displayed if missing
				if not IsKnownAbility(skill.requires) then
					if db.profile.ability_req then
						local display = SHORT_ABILITY_NAME[skill.requires] or skill.requires
						req_text = ("%s%s\124cFFFF0000%s\124r"):format(req_text, separator, display)
					end
					can_train = false
				end
				--separator = ", "
			end
			if req_text ~= "" then
				req_text = " - "..req_text
			end
			--err("Icon: %s",tostring(skill.icon))
			line, column = tooltip:AddLine()

			local line_colour = can_train and "\124cFF00FF00" or ""
				--tooltip::SetLineColor(line, r, g, b)

			text = ("%s  \124T%s:0:0\124t%s%s"):format(line_colour, skill.icon or "INV_MISC_QUESTIONMARK", skill_name, req_text)

			tooltip:SetCell(line, 1, text)

			if skill.cost then
				if db.profile.reputation then
					tooltip:SetCell(line, 2, line_colour .. skill.best_rep)
				end
				tooltip:SetCell(line, 3, _G.GetMoneyString(skill.cost), WHITE_FONT)
			end

			local skill_index = "classes$$" .. MY_CLASS_UPPER .. '$$' .. spell_id

			tooltip:SetLineScript(line, "OnEnter", ShowInfoTooltip, skill_index)
			tooltip:SetLineScript(line, "OnLeave", HideInfoTooltip)
			tooltip:SetLineScript(line, "OnMouseUp", ToggleInfoTooltip, skill_index)

		end
		ReleaseTable(sort_class_skill)
	end

	-- Display unspent talents
	if db.profile.show_unspent_talents ~= 'never' then
		local primary, secondary = ScanUnspentTalentPoints()

		-- if show_unspent_talents == 'if unspent', only show if there are point to be spent
		if db.profile.show_unspent_talents == 'always' or
			(db.profile.show_unspent_talents == 'if_unspent'  and (primary ~= 0 or (secondary and secondary ~= 0))) then

			-- Add to total for talents
			total_learnable = total_learnable + (primary ~= 0 and 1 or 0) + (secondary and secondary ~= 0 and 1 or 0)

			if not first_line then
				line, column = tooltip:AddSeparator(nil,0,0,0,0)
				line, column = tooltip:AddSeparator(nil,0,0,0,0)
			else
				first_line = nil
			end

			local text
			if secondary == nil then
				local primary_color = primary > 0 and "\124cFF00FF00" or ""
				text = ("  %s%s: %s"):format(primary_color, _G.PRIMARY, primary)
			else
				local primary_color = primary > 0 and "\124cFF00FF00" or ""
				local secondary_color = secondary > 0 and "\124cFF00FF00" or ""
				text = ("  %s%s: %s\124r, %s%s: %s"):format(primary_color, _G.PRIMARY, primary, secondary_color, _G.SECONDARY, secondary)
			end

			line, column = tooltip:AddHeader()
			tooltip:SetCell(line, 1, L["Unspent Talents"], HEADER_FONT, "LEFT")

			line, column = tooltip:AddLine()
			tooltip:SetCell(line, 1, text)

			tooltip:SetLineScript(line, "OnEnter", ShowTextTooltip, _G.BINDING_NAME_TOGGLETALENTS)
			tooltip:SetLineScript(line, "OnLeave", HideTextTooltip)
			tooltip:SetLineScript(line, "OnMouseUp", _G.ToggleTalentFrame)
		end

	end

	-- Display totals
	if total_learnable ~= 0 and db.profile.total_learnable then
		line, column = tooltip:AddLine()

		tooltip:SetCell(line, 1, " ")
		tooltip:AddSeparator()

		line, column = tooltip:AddLine()

		tooltip:SetCell(line, 1, ("\124cFFFF7F00%s\124r"):format(L['Number of skills ready to train:']))
		tooltip:SetCell(line, 3, tonumber(total_learnable), WHITE_FONT)
	end

	if total_cost ~= 0 and db.profile.overall_cost then
		line, column = tooltip:AddLine()
		tooltip:SetCell(line, 1, ("\124cFFFF7F00%s\124r"):format(L['Total costs to train the learnable skills:']))
		tooltip:SetCell(line, 3, _G.GetMoneyString(total_cost), WHITE_FONT)
	end

	-- Instructions for toggling the tooltips
	tooltip:AddSeparator()
	line, column = tooltip:AddLine()
	tooltip:SetCell(line, 1, L['Click a line to toggle between skill and trainer tooltips'], INSTRUCTION_FONT, "CENTER", 3)

	-- Table Clean-up
	ReleaseTable(class_to_display)
	ReleaseTable(prof_to_display)

	-- Adjust the hight of the tooltip so that it fits in the screen
	--tooltip:UpdateScrolling(GetScreenHeight() - 30)
	tooltip:UpdateScrolling()
	tooltip:Show()
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------
function Syllabus:PLAYER_REGEN_DISABLED()
	in_combat = true
end

function Syllabus:PLAYER_REGEN_ENABLED()
	in_combat = false

	if update_queued then
		update_queued = false
		self:UpdateSkills()
	end
end

do
	local trade_filters
	local function SavePref()
		-- Save the user's preferences before opening everything.
		-- Expend all the headers
	end

	local function RestorePref()
		-- Restore the previous values for the user's type filters.
	end

	function Syllabus:TRADE_SKILL_SHOW()
			-- We don't want to save the information if this isn't our tradeskill list.
		if _G.IsTradeSkillLinked() or _G.IsTradeSkillGuild() then
			return
		end

--		SavePref()
		_G.TradeSkillSetFilter(-1, -1)

		local prof_name = _G.GetTradeSkillLine()
		for i = 1, _G.GetNumTradeSkills() do
			local name, skill_type, _, _, _, _, _, show_progress_bar, current_rank, max_rank = _G.GetTradeSkillInfo(i)

			if name then
				if skill_type == "subheader" and show_progress_bar and EXTRA_PROFESSIONS_NAMES[name] then
					-- Update to current values
					MY_PROFESSIONS[name]					= MY_PROFESSIONS[name] or {}
					MY_PROFESSIONS[name].rank			= current_rank
					MY_PROFESSIONS[name].max_rank		= max_rank
					MY_PROFESSIONS[name].show_option	= 'show_cook'
				elseif skill_type ~= "header" and skill_type ~= "subheader" then
					local spell_id = _G.GetTradeSkillRecipeLink(i):match("enchant:(%d+)")
					spell_id = (spell_id and tonumber(spell_id)) or nil
					SetKnownSkill(spell_id, name)
				end
			end
		end

--		RestorePref()

		-- Update the broker text
		DataObj.text = GetUpdatedBrokerText()

	end
end

do
	local current_skill = {}	-- Re-usable table for skill insertions.

	local type_filters = {
		["available"] = 0,
		["unavailable"] = 0,
		["used"] = 0,
	}
	local function SavePref()
		-- Save the user's preferences before opening everything.
		for k, v in pairs(type_filters) do
			type_filters[k] = _G.GetTrainerServiceTypeFilter(k) or 0
			_G.SetTrainerServiceTypeFilter(k, 1)
		end

		-- Expend all the headers
		for i = 1, _G.GetNumTrainerServices() do
			local _ , _, category = _G.GetTrainerServiceInfo(i)

			if not category then
				break
			end

			if category == "header" then
				_G.ExpandTrainerSkillLine(i)
			end
		end
	end

	local function RestorePref()
		-- Restore the previous values for the user's type filters.
		for k, v in pairs(type_filters) do
			_G.SetTrainerServiceTypeFilter(k, type_filters[k])
		end
		_G.ClassTrainerFrame.scrollFrame.scrollBar:SetValue(0);
		_G.ClassTrainer_SelectNearestLearnableSkill()
	end

	local function FindTrainerID()
		local guid = _G.UnitGUID("npc")
		if guid == nil then return nil, nil, nil end

		local entityTypeStr, entityIdStr = guid:match("^0x(%x%x%x)%x(%x%x%x%x)")
		local entityTypeNum = tonumber("0x" .. entityTypeStr) or nil
		local entityIdNum = tonumber("0x" .. entityIdStr) or nil

		return 	entityIdNum, entityTypeNum, guid
	end

	local DataMiner = _G.CreateFrame("GameTooltip", "SyllabusDataMiner", _G.UIParent, "GameTooltipTemplate")
	DataMiner:SetOwner(_G.UIParent, "ANCHOR_NONE")

	function Syllabus:TRAINER_SHOW()
		_G.GameTooltip_SetDefaultAnchor(DataMiner, _G.UIParent)

		-- Get some information from the trainner tooltip
		DataMiner:SetUnit("npc")

		local trainer_name = _G["SyllabusDataMinerTextLeft1"]:GetText()
		local trainer_line2 = _G["SyllabusDataMinerTextLeft2"]:GetText()
		local trainer_line3 = _G["SyllabusDataMinerTextLeft3"]:GetText()
		-- Some trainers like the DK Class trainers do not have a line 3 i.e. a trainer role
		-- Let's take line 3 for reputation if there is no line 4
		local trainer_rep = _G["SyllabusDataMinerTextLeft4"]:GetText() or trainer_line3
		DataMiner:Hide()

		-- Some NPC have no reputation line but still have 3 lines
		local rep_standing_id = GetReputationStanding(trainer_rep)
		if rep_standing_id == 0 then
			trainer_rep = nil
			rep_standing_id = 4 -- We assume that there are no discount wihtout reputation
		end

		local trainer_role = nil

		_G.SetMapToCurrentZone() -- Make sure were are looking at the right zone
		local trainer_x, trainer_y = _G.GetPlayerMapPosition("player")
		--local trainer_zone = _G.GetRealZoneText()
		local trainer_subzone = _G.GetSubZoneText()
		local trainer_zone_id = _G.GetCurrentMapZone() or 0				-- 0 is for Unkown zone
		local trainer_continent_id = _G.GetCurrentMapContinent() or -1	-- 0 is for Unkown continent
		local trainer_map_id = _G.GetCurrentMapAreaID()
		local trainer_floor = _G.GetCurrentMapDungeonLevel()
		trainer_floor = trainer_floor ~= 0 and trainer_floor or nil
		local trainer_id = FindTrainerID()

		if dev then
			print(("Trainer is: %s <%s>, in %s at %.2f, %.2f. Reputation = %s, c_id = %s, z_id = %s, m_id = %s, f_id = %s, id = %s"):format(
						trainer_name,
						trainer_line2,
						ZONE_NAMES[trainer_continent_id][trainer_zone_id],
						trainer_x * 100,
						trainer_y * 100,
						trainer_rep or 'nil',
						tostring(trainer_continent_id),
						tostring(trainer_zone_id),
						tostring(trainer_map_id),
						tostring(trainer_floor),
						tostring(trainer_id)
			))
		end

		SavePref()

		-- Get the data from the open Trainer Windows
		local trainer_db
		local is_tradeskill = false
		-- Special fix now that Riding is learned from a Trade Skill Trainer
		local is_riding = (_G.GetTrainerServiceSkillLine(1) == L['Riding'])

		if _G.IsTradeskillTrainer() and not is_riding then
			trainer_db = skills_db.professions
			is_tradeskill = true
		else
			trainer_db = skills_db.classes
		end

		for idx = 1, _G.GetNumTrainerServices() do
			local svc_name, _, svc_type, svc_texture, svc_reqlevel = _G.GetTrainerServiceInfo(idx)

			if svc_name and svc_type and svc_type ~= "header" then
				local level_req	= _G.GetTrainerServiceLevelReq(idx)
				local skill_req, rank_req, _ = _G.GetTrainerServiceSkillReq(idx)
				local spell_id

				if rank_req == 0 then
					skill_req = nil
					rank_req = nil
				end
				local num_reqs = _G.GetTrainerServiceNumAbilityReq(idx)
				local ability_reqs, meets_reqs

				if num_reqs > 0 then
					ability_reqs, meets_reqs = _G.GetTrainerServiceAbilityReq(idx, 1)
					-- Save the ability for the toon
					if meets_reqs then
						SetKnownSkill(ABILITY_NAME_TO_SPELLID[ability_reqs], ability_reqs)
					end
				end

				--local item_link = _G.GetTrainerServiceItemLink(idx)

				--current_skill.item_link = item_link and item_link:match("item:(%d+)") or nil
				current_skill.item_link = nil -- remove old debug value
				current_skill.level_req = ((level_req > 0) and level_req) or nil
				current_skill.cost = GetBaseCost(_G.GetTrainerServiceCost(idx), rep_standing_id)
				current_skill.rank_req = rank_req
				current_skill.requires = ability_reqs and ABILITY_NAME_TO_SPELLID[ability_reqs] or ability_reqs
				current_skill.num_reqs = num_reqs
				--current_skill.svc_header = _G.GetTrainerServiceSkillLine(idx)
				current_skill.svc_header = nil -- remove old debug value
				current_skill.name = svc_name
				current_skill[MY_FACTION_FLAG] = true

				if dev and type(current_skill.requires) == "string" then
					err("Missing spell in IS_ABILITY: %s", current_skill.requires)
				end

				-- We mine the Tooltip again to get the SpellID
				_G.GameTooltip_SetDefaultAnchor(DataMiner, _G.UIParent)
				DataMiner:SetTrainerService(idx)
				spell_id = select(3,DataMiner:GetSpell())
				DataMiner:Hide()

				-- Only keep the texture if it is different from the spell texture
				current_skill.icon = svc_texture ~= _G.GetSpellTexture(spell_id) and svc_texture or nil

				local svc_header
				if is_tradeskill then
					svc_header = _G.GetTrainerServiceSkillLine(idx)
				else
					svc_header = MY_CLASS_UPPER
				end

				-- Keep the first svc_name for the trainer
				if not trainer_role then
					trainer_role = svc_header
				elseif trainer_role ~= svc_header then
					err("trainer_role changed from %s to %s",trainer_role,svc_header)

				end

				-- Overwrite existing data with what was found but nothing is ereased
				for key, val in pairs(current_skill) do
					trainer_db[svc_header][spell_id][key] = val
				end

				if trainer_rep then
					trainer_db[svc_header][spell_id].valid_rep[trainer_rep] = true
				end
				trainer_db[svc_header][spell_id].trainers[trainer_id] = true
				-- Remove the trainer name now that we know the id
				trainer_db[svc_header][spell_id].trainers[trainer_name] = nil

				-- Identify riding skills and save the data to be shared between all classes
				if RIDING_SKILLS[spell_id] then
					for key, val in pairs(current_skill) do
						skills_db.all_classes[spell_id][key] = val
					end

					if trainer_rep then
						skills_db.all_classes[spell_id].valid_rep[trainer_rep] = true
					end
					skills_db.all_classes[spell_id].trainers[trainer_id] = true
					-- Remove the trainer name now that we know the id
					skills_db.all_classes[spell_id].trainers[trainer_name] = nil
				end

				if svc_type == "used" and svc_name ~= svc_header then
					SetKnownSkill(spell_id, svc_name)
				end
			end
		end

		-- If there is an entry my name, convert to entry by ID
		if not IsTableEmpty(trainers_db[trainer_name]) then
			trainers_db[trainer_id] = trainers_db[trainer_name]
			trainers_db[trainer_name] = nil
		end

		local trainer_info = trainers_db[trainer_id]
		-- Sometimes, GetPlayerMapPosition doesn't return valid coordinates
		-- I do not know why yet. trainer_x is the only one tested on purpuse.
		trainer_info.name = trainer_name
		trainer_info.x = (trainer_x ~= 0 and trainer_x) or trainer_info.x
		trainer_info.y = (trainer_x ~= 0 and trainer_y) or trainer_info.y
		--trainer_info.zone = trainer_zone
		trainer_info.subzone = trainer_subzone
		trainer_info.zone_id =  trainer_zone_id
		trainer_info.continent_id =  trainer_continent_id
		trainer_info.map_id = trainer_map_id
		trainer_info.floor = trainer_floor
		trainer_info.id = trainer_id
		--trainer_info.line2 = trainer_line2 or ""
		--trainer_info.line3 = trainer_line3 or ""
		trainer_info.line2 = nil -- Remove old debug values
		trainer_info.line3 = nil -- Remove old debug values
		trainer_info.rep = trainer_rep
		trainer_info[MY_FACTION_FLAG] = true

		RestorePref()
		Syllabus:UpdateSkills()
	end

	function Syllabus:TRAINER_CLOSED()
		--err("TRAINER_CLOSED called")
		-- On close, we just recheck the values on trained skills
		SavePref()

		--local total, total_known = 0, 0
		for idx = 1, _G.GetNumTrainerServices() do
			local header
			local name, subType, category, texture, requiredLevel, topServiceLine = _G.GetTrainerServiceInfo(idx)

			if _G.IsTradeskillTrainer() then
				header = _G.GetTrainerServiceSkillLine(idx)
			else
				header = MY_CLASS_UPPER
			end

			-- We mine the Tooltip to get the SpellID
			_G.GameTooltip_SetDefaultAnchor(DataMiner, _G.UIParent)
			DataMiner:SetTrainerService(idx)
			local spell_id = select(3,DataMiner:GetSpell())
			DataMiner:Hide()

			if name and category and category == "used" and name ~= header then
				SetKnownSkill(spell_id, name)
			end
		end

		--err("TRAINER_CLOSED: total = %s, total_known = %s", total, total_known)

		RestorePref()
		self:UpdateSkills()
	end
end	-- do

-- Clean up the SaveVariables
function Syllabus:CleanCummonClassSpells()
	local skills_db = Syllabus.db.global.skills

	-- Protection againsts a weird case when doing /reload
	if not skills_db or not skills_db["all_classes"] or not skills_db["classes"] then return end

	local is_commun_spell = {}
	local all_classes = skills_db["all_classes"]
	local classes = skills_db["classes"]

	for spell_id in pairs(all_classes) do
		is_commun_spell[spell_id] = true
	end

	for class, class_spells in pairs(classes) do
		for spell_id, spell in pairs(class_spells) do
			if is_commun_spell[spell_id] and not IsTableEmpty(spell) then
				wipe(spell)
				class_spells[spell_id] = nil
			end
		end
	end
end

-- Big hack!!!!!
-- Build localized: "^Your skill in (.*) has increased to (%d+).$"
local SKILL_UP_MSG	= _G.ERR_SKILL_UP_SI:gsub("%%d",'(%%d+)')
SKILL_UP_MSG 			= '^' .. SKILL_UP_MSG:gsub("%%s",'(.*)') .. '$'

function Syllabus:CHAT_MSG_SKILL(event, msg, format)
	local skill_name, rank = msg:match(SKILL_UP_MSG)
--err("msg = %s, skill_name= %s, rank= %s", msg, skill_name, rank)
	if skill_name and EXTRA_PROFESSIONS_NAMES[skill_name] then
		MY_PROFESSIONS[skill_name] = MY_PROFESSIONS[skill_name] or {}
		MY_PROFESSIONS[skill_name].rank = tonumber(rank)
		MY_PROFESSIONS[skill_name].max_rank = MY_PROFESSIONS[skill_name].max_rank or 600
		MY_PROFESSIONS[skill_name].show_option = 'show_cook'
		self:UpdateSkills()
	end
end

function Syllabus:PLAYER_LOGOUT()
	-- Remove the riding spells from the classes
	self:CleanCummonClassSpells()
	
	-- Save the cooking specialisation professions
	db.char.cooking_prof = {}
	for prof_name in pairs(MY_PROFESSIONS) do
		if EXTRA_PROFESSIONS_NAMES[prof_name] then
			db.char.cooking_prof[prof_name] 					= {}
			db.char.cooking_prof[prof_name].rank 			= MY_PROFESSIONS[prof_name].rank
			db.char.cooking_prof[prof_name].max_rank 		= MY_PROFESSIONS[prof_name].max_rank
			db.char.cooking_prof[prof_name].show_option	= MY_PROFESSIONS[prof_name].show_option
		end
	end
end

-------------------------------------------------------------------------------
-- Data convertion
-------------------------------------------------------------------------------

-- This function convert (or reset) data when the Syllabus data model is changed
-- or when a major event in game make it mandantory.
-- It is called before AceDB is bound to the SaveVariable.
-- The current version of the data model is found in .global.version
function Syllabus:ConvertData()

	local db = _G[MODNAME.."DB"]
	local bliz_build = select(2,_G.GetBuildInfo())
	bliz_build = tonumber(bliz_build)

	-- If there are no variables yet
	if not db or not db.global then
		_G[MODNAME.."DB"] = _G[MODNAME.."DB"] or {}
		_G[MODNAME.."DB"].global = _G[MODNAME.."DB"].global or {}
		_G[MODNAME.."DB"].global.version = IS_504 and 1.0 or 0.9
		return
	end

	-- For Mists, we need to reset the database. To many things have changed
	if IS_504 and db.global.version and db.global.version < 1.0 then

		db.char = nil
		db.global.trainers = nil
		db.global.skills = nil

		_G.message(L.ERASE_DATA_504_MESSAGE)

		db.global.version = 1.0
	end
	
	if db.global.version == 1.0 then
	
		-- Remove old fishing skills because they changed ID and generate errors
		local fishing = _G.GetSpellInfo(131474) or "Fishing"
		local fishing_db = db.global.skills.professions[fishing]
		
		fishing_db[7620] = nil
		fishing_db[7731] = nil
		fishing_db[7732] = nil
		fishing_db[18248] = nil
		fishing_db[33095] = nil
		fishing_db[51294] = nil
		fishing_db[88868] = nil
		
		db.global.version = 1.1
	end

end

-------------------------------------------------------------------------------
-- Initialization functions
-------------------------------------------------------------------------------

function Syllabus:OnInitialize()
	-- Defaults describe the data structure
	local defaults = {
		global = {
			version = nil,
			trainers = {
				-- Trainer guid
				['*'] = {},
			},
			skills = {
				classes = {
					-- Class name
					['*'] = {
						-- Spell ID
						['*'] = {
							valid_rep = {},
							trainers = {},
						},
					},
				},
				-- Skills that are shared between all classes (Riding for now)
				all_classes = {
					-- Spell ID
					['*'] = {
						valid_rep = {},
						trainers = {},
					},
				},
				professions = {
					['*'] = {
						-- Spell ID
						['*'] = {
							valid_rep = {},
							trainers = {},
						},
					},
				},
			},
		},
		char = {
			known = {
				-- SpellID = something, I use the spell name for ease of debugging
				['*'] = nil,
			},
			cooking_prof = { },
		},
		profile = {
			-- Configuration data
			-- Data Broker display
			broker_learnable = true,
			-- cost = none, base or cheapest
			broker_cost = 'cheapest',
			broker_filter_cost = 0,
			broker_nothing_msg = "0",

			-- Main Tooltip display
			reputation = false,
			display_cost = 'cheapest',
			filter_cost = 0,
			total_learnable = true,
			overall_cost = true,
			level_range = 5,
			skill_range = 25,
			ability_req = true,
			tooltip_scale = 1,

			-- Professions
			show_arch = true,
			show_cook = true,
			show_firstaid = true,
			show_fish = true,
			show_riding = true,
			show_prof1 = true,
			show_prof2 = true,

			show_class = true,

			show_unspent_talents = 'if_unspent',

			not_trainable = false,
			requirements = true,
			skill_count = true,
			trainable = true,
			trainable_future = true,
			trainer_names = true,
			skill_tooltip = true,
			trainer_by_subzone = 1,
		}
	}

	-- Data convertion before SyllabusDB is bound to AceDB
	self:ConvertData()

	db = _G.LibStub("AceDB-3.0"):New(MODNAME.."DB", defaults, true)
	Syllabus.db = db
	known_db = db.char.known
	trainers_db = db.global.trainers
	skills_db = db.global.skills
	
	-- Load the saved cooking speciality professions
	for prof_name in pairs(EXTRA_PROFESSIONS_NAMES) do
		if db.char.cooking_prof[prof_name] then
			MY_PROFESSIONS[prof_name]					= db.char.cooking_prof[prof_name]
			MY_PROFESSIONS[prof_name].rank			= db.char.cooking_prof[prof_name].rank
			MY_PROFESSIONS[prof_name].max_rank		= db.char.cooking_prof[prof_name].max_rank
			MY_PROFESSIONS[prof_name].show_option	= db.char.cooking_prof[prof_name].show_option
		end
	end

	-- Data broker
	DataObj = LDB:NewDataObject(MODNAME, {
		type	= "data source",
		label	= MODNAME,
		text	= db.broker_nothing_msg,
		icon	= "Interface\\Icons\\INV_Misc_Book_01",
		SetIcon	= function(self, icon)
				  self.icon = icon or "Interface\\Icons\\INV_Misc_Book_01"
				end,
		OnEnter	= function(display, motion)
				  ShowTooltip(display)
				end,
		OnLeave	= function(display, motion)
				end,
		OnClick  = function(display, button, down)
					-- Only when the button is release
					if down then return end

					-- For tests of EasyMenu
					if button ==  "RightButton" then
						_G.InterfaceOptionsFrame_OpenToCategory(MODNAME)
					end
				end,
	})

	self:SetupOptions()
	self.OnInitialize = nil
end

function Syllabus:OnEnable()
	self:InitKnownspells()
	self:UpdateSkills()
	self:RegisterEvent("PLAYER_LEVEL_UP", "UpdateSkills")
	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:RegisterEvent("PLAYER_TALENT_UPDATE", "UpdateSkills")
	self:RegisterEvent("SKILL_LINES_CHANGED", "UpdateSkills")
	self:RegisterEvent("TRADE_SKILL_SHOW")
	self:RegisterEvent("TRAINER_SHOW")
	self:RegisterEvent("TRAINER_CLOSED")
	self:RegisterEvent("CHAT_MSG_SKILL")
	self:RegisterEvent("PLAYER_LOGOUT")
end

function Syllabus:OnDisable()
	self:UnregisterEvent("PLAYER_LEVEL_UP")
	self:UnregisterEvent("PLAYER_REGEN_DISABLED")
	self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	self:UnregisterEvent("PLAYER_TALENT_UPDATE")
	self:UnregisterEvent("SKILL_LINES_CHANGED")
	self:UnregisterEvent("TRADE_SKILL_SHOW")
	self:UnregisterEvent("TRAINER_SHOW")
	self:UnregisterEvent("TRAINER_CLOSED")
	self:UnregisterEvent("CHAT_MSG_SKILL")
	self:UnregisterEvent("PLAYER_LOGOUT")
end

-------------------------------------------------------------------------------
-- Local variables and functions that are used in other files
-------------------------------------------------------------------------------

-- Constants
Syllabus.ABILITY_NAME_TO_SPELLID	= ABILITY_NAME_TO_SPELLID
Syllabus.MY_CLASS_UPPER				= MY_CLASS_UPPER
Syllabus.MY_FACTION_FLAG			= MY_FACTION_FLAG
Syllabus.ZONE_NAMES					= ZONE_NAMES
Syllabus.MAPID_TO_NAME				= MAPID_TO_NAME

-- Functions
Syllabus.CopyTable					= CopyTable
Syllabus.IsTableEmpty				= IsTableEmpty
Syllabus.GetStaticProfName			= GetStaticProfName
Syllabus.GetLocalProfName			= GetLocalProfName

-------------------------------------------------------------------------------
-- TEST: Level up messages
-------------------------------------------------------------------------------

local levelUpTexCoords = {
	[_G.LEVEL_UP_TYPE_CHARACTER] = {
		dot = { 0.64257813, 0.68359375, 0.18750000, 0.23046875 },
		goldBG = { 0.56054688, 0.99609375, 0.24218750, 0.46679688 },
		gLine = { 0.00195313, 0.81835938, 0.01953125, 0.03320313 },
	},
}

function Syllabus:SkillUpDisplay(skill, skill_level)
	local LevelUpDisplay = _G.LevelUpDisplay

	LevelUpDisplay.type = _G.LEVEL_UP_TYPE_CHARACTER

	_G.LevelUpDisplay_BuildCharacterList(LevelUpDisplay);
	LevelUpDisplay.levelFrame.reachedText:SetText(_G.LEVEL_UP_YOU_REACHED)
	LevelUpDisplay.levelFrame.levelText:SetFormattedText("",skill, skill_level);

	LevelUpDisplay.gLine:SetTexCoord(unpack(levelUpTexCoords[LevelUpDisplay.type].gLine));
	LevelUpDisplay.gLine2:SetTexCoord(unpack(levelUpTexCoords[LevelUpDisplay.type].gLine));
	if (levelUpTexCoords[LevelUpDisplay.type].tint) then
		LevelUpDisplay.gLine:SetVertexColor(unpack(levelUpTexCoords[LevelUpDisplay.type].tint));
		LevelUpDisplay.gLine2:SetVertexColor(unpack(levelUpTexCoords[LevelUpDisplay.type].tint));
	else
		LevelUpDisplay.gLine:SetVertexColor(1, 1, 1);
		LevelUpDisplay.gLine2:SetVertexColor(1, 1, 1);
	end

	if (levelUpTexCoords[LevelUpDisplay.type].textTint) then
		LevelUpDisplay.levelFrame.levelText:SetTextColor(unpack(levelUpTexCoords[LevelUpDisplay.type].textTint));
	else
		LevelUpDisplay.levelFrame.levelText:SetTextColor(_G.NORMAL_FONT_COLOR.r, _G.NORMAL_FONT_COLOR.g, _G.NORMAL_FONT_COLOR.b);
	end

	LevelUpDisplay.levelFrame.levelUp:Play();

end


--[=[
	function me.TestID ( NpcID )
		Tooltip:SetOwner( WorldFrame, "ANCHOR_NONE" );
		Tooltip:SetHyperlink( ( "unit:0xF53%05X00000000" ):format( NpcID ) );
		if ( Tooltip:IsShown() ) then
			return Text:GetText();
		end
	end

]=]